        module ParseLib where


infixl 6  <*>, <$>, <*, *>, <$
infixr 4  <|>

--                    

type Parser symbol result  =  [symbol] -> [(result,[symbol])]

--                   

--                        -            
symbol  ::  (s -> s -> Bool) -> s -> Parser s s
symbol _ _ []                  =  []
symbol eq a (x:xs) | x `eq` a     =  [(x,xs)]
                | otherwise  =  []

satisfy  ::  (s -> Bool) -> Parser s s
satisfy _ []                 =  []
satisfy p (x:xs) | p x       =  [(x,xs)]
                 | otherwise =  []

--                       -              
token  ::  (s -> s -> Bool) -> [s] -> Parser s [s]
token eq k xs  |  eqList eq k (take n xs)  =  [(k,drop n xs)]
            |  otherwise       =  []
  where  n = length k

failp     :: Parser s a
failp _  =  []

succeed       :: a -> Parser s a
succeed r xs  =  [(r,xs)]

--                                   

digit  :: Parser Char Char
digit  =  satisfy isDigit

--                    

(<|>)         :: Parser s a      -> Parser s a -> Parser s a
(p <|> q) xs  =  p xs ++ q xs

(<*>)         :: Parser s (b -> a) -> Parser s b -> Parser s a
(p <*> q) xs  =  [(f x,zs)
                 |(f  ,ys) <- p xs
                 ,(  x,zs) <- q ys
                 ]

(<$>)         :: (a -> b) -> Parser s a -> Parser s b
(f <$> p) =  succeed f <*> p

--                   

(<*) :: Parser s a -> Parser s a -> Parser s a
p <* q    = const <$> p <*> q

(*>) :: Parser s a -> Parser s a -> Parser s a
p *> q    = flip const <$> p <*> q

(<$) :: b -> Parser s a -> Parser s b
f <$ p    = const f <$> p

--                                   

newdigit  :: Parser Char Int
newdigit  =  f <$> digit
  where  f c = ord c - ord '0'

--                         

option      :: Parser s a -> a -> Parser s a
option p d  =  p <|> succeed d

many    :: Parser s a  -> Parser s [a]
many p  =  list <$> p <*> many p <|> succeed []

many1    :: Parser s a -> Parser s [a]
many1 p  =  list <$> p <*> many p

pack        :: Parser s a -> Parser s b -> Parser s c -> Parser s b
pack p r q  =  pi32 <$> p <*> r <*> q 

listOf      :: Parser s a -> Parser s b -> Parser s [a]
listOf p s  =  list <$> p <*> many (pi22 <$> s <*> p)

--                    

determ  ::  Parser s b -> Parser s b
determ p xs  |  null r     =  []
             |  otherwise  =  [head r]
  where r = p xs

greedy, greedy1  ::  Parser s b -> Parser s [b]
greedy   =  determ . many
greedy1  =  determ . many1

list :: a -> [a] -> [a]
list x xs  =  x:xs

pi22 :: a -> b -> b
pi22 _ y    =  y
pi32 :: a -> b -> c -> b
pi32 _ y _  =  y

--                                 

natural  :: Parser Char Int
natural  =  foldl (\a b -> a*10 + b) 0 <$> many1 newdigit

integer  ::  Parser Char Int
integer  =  (const negate <$> (symbol eqChar '-')) `option` id  <*>  natural

identifier :: Parser Char String
identifier =  list <$> satisfy isAlpha <*> greedy (satisfy isAlphaNum)

parenthesised :: Parser Char a -> Parser Char a
parenthesised p  =  pack (symbol eqChar '(') p (symbol eqChar ')')

commaList    :: Parser Char a -> Parser Char [a]
commaList p  =  listOf p (symbol eqChar ',')

--                             

chainr  ::  Parser s a -> Parser s (a -> a -> a) -> Parser s a
chainr pe po  =  h <$> many (j <$> pe <*> po) <*> pe
  where j x op  =  (x `op`)
        h fs x  =  foldr ($) x fs

chainl  ::  Parser s a -> Parser s (a -> a -> a) -> Parser s a
chainl pe po  =  h <$> pe <*> many (j <$> po <*> pe)
  where j op x  =  (`op` x)
        h x fs  =  foldl (flip ($)) x fs


--                                           

psequence         :: [Parser s a] -> Parser s [a]
psequence []      =  succeed []
psequence (p:ps)  =  list <$> p <*> psequence ps

psequence'  :: [Parser s a] -> Parser s [a]
psequence'  =  foldr f (succeed [])
  where  f p q = list <$> p <*> q

choice  :: [Parser s a] -> Parser s a
choice  =  foldr (<|>) failp


