module Grouping (
        module Grouping
      , group
      , groupBy
      )
   where

-------------------------------------------------------
-- Grouping utilities 
-------------------------------------------------------

import Data.List (group, groupBy, sort, sortBy, partition, delete)
import qualified Data.Map as Map

import Auxilary

{-
from Haskell Hierarchical Libraries (base package) Data.List (http://www.haskell.org/ghc/docs/latest/html/libraries/base/Data-List.html):

group :: Eq a => [a] -> [[a]]

The group function takes a list and returns a list of lists such that the concatenation of the result is equal to the argument. Moreover, each sublist in the result contains only equal elements. For example,

 group "Mississippi" = ["M","i","ss","i","ss","i","pp","i"]

It is a special case of groupBy, which allows the programmer to supply their own equality test. 

groupBy :: (a -> a -> Bool) -> [a] -> [[a]]
The groupBy function is the non-overloaded version of group.

-}

----
-- group only groups sequencing elements, from Eq class
-- some usefull variants:
----

-- Grouping all elements with standard ordering, result is sorted
groupAll :: Ord a => [a] -> [[a]]
groupAll =  group . sort

-- Grouping all elements with custom ordering, result is sorted
groupAllByOrd :: (a -> a -> Ordering) -> [a] -> [[a]]
groupAllByOrd ord = groupBy eq . (sortBy ord)
   where
      eq a b =
         case ord a b of
            EQ -> True
            GT -> False
            LT -> False


{-

Grouping> groupAllByEq' (\x y -> even x && even y) [0..5]
[[0,2,4]]
instead of: [[0,2,4],[1],[3],[5]]


doesnt work properly, 

-- similar to groupBy, but now also groups non following elements
groupAllByEq :: (a -> a -> Bool) -> [a] -> [[a]]
groupAllByEq eqPred set = filter (not . null) $ splitSet selectFuns set
   where
      selectFuns = map eqPred set

-- a group Under Set of functions
splitSet :: [a -> Bool] -> [a] -> [[a]]
splitSet []          _     = []
splitSet (fun:funs)  set   = (fst g) : splitSet funs (snd g)
   where g = partition fun set

-- can also be implemented with a fold probl., but then the transforming function has to keep track of the sets diveded over the different groups of sets
-- exampl : blocks blockSize = takeWhile ( not . null ) . map fst . iterate (splitAt blockSize . snd) . splitAt blockSize
-}

groupAllByEq :: (a -> a -> Bool) -> [a] -> [[a]]
groupAllByEq = groupByCoherence  -- using grouping as used with loggings sessions

-- group all by order preserving by arie, using Map for O(log n) insertion
groupAllBy_Map :: Ord b => (a -> b) -> [a] -> [[a]]
groupAllBy_Map f = Map.elems . foldr (\v -> Map.insertWith (++) (f v) [v]) Map.empty

{-
groupAllBy_Map' :: (a -> a -> Bool) -> [a] -> [[a]]
groupAllBy_Map' f = groupAllBy_Map f
-}

newtype Sequence a = Sequence [a] -- all elements are sequencially ordered 

-- Grouping a sequence of (multi-dimensional) elements according to a isSequence / coherence predicate. The input collection of elements should be a/in sequence, for loggings this means that they should following each other in time (temperol ordered). Simple values like numbers are not multi-dimensional but can easily be grouped similarly like groupAllByEq

{-
groupByCoherence :: Eq a => (a -> a -> Bool) -> {-Sequence -} [a] -> [[a]]
groupByCoherence cohPred list =
   let
      lift :: (a -> a -> Bool) -> ([a] -> [a] -> Bool)
      lift _    []    _     = error "Empty list when lifting predicate in grouping by coherence function"
      lift _    _     []    = error "Empty list when lifting predicate in grouping by coherence function"
      lift pred list1 list2 = pred (last list1) (head list2)

      insertGroupBy   :: Eq a => ([a] -> [a] -> Bool) -> [a] -> [[a]] -> [[a]]
      insertGroupBy _    grp []     = [grp]
      insertGroupBy pred grp groups =
         case (length coh_grps > 0) of
            True -> (grp ++ (head coh_grps)) : ( delete (head coh_grps) groups )
            False -> grp : groups
         where
            coh_grps = filter (pred grp) groups
   in
      foldr (insertGroupBy $ lift cohPred) [] (groupBy cohPred list)
-}
groupByCoherence :: (a -> a -> Bool) -> {-Sequence -} [a] -> [[a]]
groupByCoherence cohPred list =
   let
      lift :: (a -> a -> Bool) -> ([a] -> [a] -> Bool)
      lift _    []    _     = error "Empty list when lifting predicate in grouping by coherence function"
      lift _    _     []    = error "Empty list when lifting predicate in grouping by coherence function"
      lift pred list1 list2 = pred (last list1) (head list2)

      insertGroupBy   :: {- Eq a => -} ([a] -> [a] -> Bool) -> [a] -> [[a]] -> [[a]]
      insertGroupBy _    grp []     = [grp]
      insertGroupBy pred grp groups =
         case (length coh_grps > 0) of
            -- True -> (grp ++ (head coh_grps)) : ( delete (head coh_grps) groups )
            -- to avoid the extra Eq a context req, use:
            True -> (grp ++ (head coh_grps)) : (pregrps ++ lastgrps)
            False -> grp : groups
         where
            -- coh_grps = filter (pred grp) groups
            (pregrps, coh_grps, lastgrps) = span3 (pred grp) groups
   in
      foldr (insertGroupBy $ lift cohPred) [] (groupBy cohPred list)

-- Order preserving span3, splits input in 3 tuple ([false elm], [fst true elm], [rest_elm])
-- > span3 even [3,1,5,6,8,9,10,11]
-- ([3,1,5],[6],[8,9,10,11])
span3 :: (a -> Bool) -> [a] -> ([a],[a],[a])
span3 pred list = (false_elms,
                   take 1 true_elms {- take returns xs if n>length xs -},
                   drop 1 true_elms {- drop returns [] if n>length xs -}
                  )
   where (false_elms, true_elms) = span (not . pred) list

groupSequence :: Eq a => (a -> a -> Bool) -> {-Sequence -} [a] -> [[a]]
groupSequence = groupByCoherence 

{-
-- groupPer :: (Logging->Logging->Bool) -> [Logging] -> [[Logging]]
groupPer :: (a -> a -> Ordering) -> [a] -> [[a]]
groupPer = groupAllByOrd -- based on groupBy . sortBy
-}
groupSimilar :: Similar a => [a] -> [[a]]
groupSimilar = groupAllByEq similar
   
-- groupOrdSimilar :: OrdSimilar a => [a] -> [[a]]
-- groupOrdSimilar = groupAllByOrd ordSimilar

--Other fun's
-- from John Meacham
-- | group a list based on a function of the values.
groupUnder :: Eq b => (a -> b) -> [a] -> [[a]]
groupUnder f = groupBy (\x y -> f x == f y)



-- disabled because groupAllByEq doesnt seem to work properly, see groupAllByEq

groupAllUnder :: Eq b => (a -> b) -> [a] -> [[a]]
groupAllUnder fun =
   let eq a b = (fun a) == (fun b)
   in
      groupAllByEq eq

groupAllUnderAndTuple :: Eq b => (a -> b) -> [a] -> [(b, [a])]
groupAllUnderAndTuple fun =
   map (\res -> (fun $ head res, res)) . groupAllUnder fun

groupAllUnderAndTuple2 :: Ord b => (a -> b) -> [a] -> [(b, [a])]
groupAllUnderAndTuple2 fun =
   enTupleGroup fun . groupAllUnder fun

-- group2Tuple :: Eq a => [(a,b)] -> [[(a, b)]]
-- group2Tuple = undefined

group3Tuple :: Eq a => [(a,b,c)] -> [(a, [(b, c)])]
group3Tuple = map
               (\grp -> (fst3 $ head grp, map (\(fstelm, sndelm, thdelm) -> (sndelm, thdelm)) grp))
               . (groupAllUnder fst3)

-- to Auxilary?
enTupleGroup :: (a -> b) -> [[a]] -> [(b,[a])]
enTupleGroup fun set =
   map (\res -> (fun $ head res, res)) set

-- Grouping Similar items
--import Grouping.Similar

-- for simply grouping on a datatype
class Eq a => Similar a where
   similar:: a -> a -> Bool

-- for grouping and sorting on a datatype
class Similar a => OrdSimilar a where
   ordSimilar:: a -> a -> Ordering

instance Similar a => Similar [a] where
   similar [] _ = False
   similar _ [] = False
   similar (x:xs) (y:ys) = similar x y && similar xs ys

{-

instance OrdSimilar a => OrdSimilar [a] where
   ordSimilar [] _ = False
   ordSimilar _ [] = False
   ordSimilar (x:xs) (y:ys) = ordSimilar x y && ordSimilar xs ys

-}
