module Statistics.DescriptiveAnalysis.DescriptiveKeyAC (
   module Statistics.DescriptiveAnalysis.DescriptiveKeyAC
   , Statistics.DescriptiveAnalysis.KeyableAC.AnalysisFK
   , Statistics.DescriptiveAnalysis.KeyableAC.Keyable
   , Statistics.DescriptiveAnalysis.KeyableAC.combine
   , Statistics.DescriptiveAnalysis.KeyableAC.startkey
   , Statistics.DescriptiveAnalysis.KeyableAC.runAnalysis
   , Statistics.DescriptiveAnalysis.KeyableAC.getAnaResult
   , Statistics.DescriptiveAnalysis.KeyableAC.peekKey
   , (Statistics.DescriptiveAnalysis.PrimAC.<.>)
   , (Statistics.DescriptiveAnalysis.PrimAC.<$>)
   ,  Statistics.DescriptiveAnalysis.PrimAC.kickStartAnalysis
   ,  Statistics.DescriptiveAnalysis.PrimAC.multiAnalyses
   ,  Statistics.DescriptiveAnalysis.PrimAC.Analysis
   ,  Statistics.DescriptiveAnalysis.PrimAC.mapAnalysis     -- helper function
   ,  Statistics.DescriptiveAnalysis.PrimAC.isolateAnalysis -- helper function
   ,  Statistics.DescriptiveAnalysis.PrimAC.splitAnalysis   -- helper function
   , (Statistics.DescriptiveAnalysis.PrimAC.<..>)           -- helper function
   ) where


import Statistics.DescriptiveAnalysis.PrimAC ((<.>), (<$>), runAnalysis, multiAnalyses, mapAnalysis, kickStartAnalysis)
import qualified Statistics.DescriptiveAnalysis.PrimAC

import Statistics.DescriptiveAnalysis.KeyableAC ( AnalysisFK
                                      , Keyable, combine, startkey
                                      , runAnalysis, getAnaResult
                                      , peekKey
                                      )
import qualified Statistics.DescriptiveAnalysis.KeyableAC

import Statistics.BasicTypes

-- for examples
{-
import Thesis.Helium.Logging
import Thesis.Grouping.Logging
import Thesis.Select
import Statistics.SummaryMeasures
import Auxilary
-}

-- To have more options for data presentations/visualisations, a special key data type
class (Keyable a {-key of previous keys in analysis sequence-} ) => DescriptiveKey a where
   basicKey :: String -> a
   aggregateKey :: String -> [a] -> a

   -- optional the actual grouping value stored in the key, supported with displaying info in the DataInfo type class
   groupKey :: DataInfo b => Either b (String, String) -> a 
   selectGrpKey :: String -> a
   tupleKey :: a -> a -> a -> a

{-
   -- Variants (strongly influence the combinators):
   basicAnakey  :: DataInfo a => b -> String -> a

   -- also storing the value using DataInfo, limited use, only members from DataInfo can be used in the resulted key
   selectGrpKey :: DataInfo b => Maybe b -> String -> a

   -- keyable like grouping, description / type (ex. identity) , actual value (ex. group23)
   groupingkey  :: String -> String -> a     

   -- the actual grouping value stored in the key, supported with displaying info in the DataInfo type class, combinators for this are in older versions of this file
   groupingkey  :: DataInfo b => Maybe b -> String -> String -> a

   -- properties of the DataInfo class tupled in the key, giving the flexibility to specify this with special grouping combinator. This is also possible in a previous for by tagging the datatype and provinding a new DataInfo instance for the new datatype
   groupingkey  :: Maybe (b, b -> AxisDataInfo b) -> String -> String -> a
   
   -- description, list of keys from inner/lifted analysis (can be a grouping-like)
   mappingkey   :: Keyable b => String -> [b] -> a 
   mappingkey   :: String -> [a] -> a -- description, list of keys from inner/lifted analysis (can be a grouping-like), could be a different type then the one of the new (outer) analysis, but this requires a polymorfic Key type
-}

-- runAnalysis from from Keyable.hs

basicAnalysis' :: DescriptiveKey key => String -> (a -> b) -> AnalysisFK key a b
basicAnalysis' keydescr ana =
   Statistics.DescriptiveAnalysis.KeyableAC.basicAnalysis (basicKey keydescr) ana

-- no annotation at all, usefull for small transformations, ex. taking head from singleton list
basicAnalysis_ :: DescriptiveKey key => (a -> b) -> AnalysisFK key a b
basicAnalysis_ ana =
   Statistics.DescriptiveAnalysis.KeyableAC.basicAnalysis (basicKey "") ana

basicAnalysis          ::  DescriptiveKey key => String -> (a -> b) -> AnalysisFK key a b
basicAnalysis          =   basicAnalysis'

----
-- Grouping Combinators with using extra DataInfo on the grouped subject
-- using these without an key which supports them doesn't cause any problems
----

groupAnalysis'' :: (DescriptiveKey key, DataInfo b) =>  ([[a]] -> [b]) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis'' keyfun grpfun =
   Statistics.DescriptiveAnalysis.KeyableAC.groupAnalysis''
      (\grps -> map (\realvalue -> groupKey (Left realvalue)) $ keyfun grps)
      grpfun

groupAnalysis'  :: (DescriptiveKey key, DataInfo b) => ([a] -> b) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis' keydescrfun grpfun =
   groupAnalysis''
      (map keydescrfun)
      grpfun

groupAnalysis          ::  (DescriptiveKey key, DataInfo b)  =>  (a  -> b) ->
                                                                 ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis keydescr grpfun = groupAnalysis' (keydescr . head) grpfun

groupAnalysisSilent :: DescriptiveKey key => ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysisSilent = Statistics.DescriptiveAnalysis.PrimAC.groupAnalysisSilent

----
-- Grouping Combinators without extra DataInfo on the grouped subject ( _ = simple )
---- 
groupAnalysis''_  :: DescriptiveKey key => ([[a]] -> [(String, String)]) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis''_ keyfun grpfun =
   Statistics.DescriptiveAnalysis.KeyableAC.groupAnalysis''
      (\grps -> map (\nonDataInfodescr -> groupKey (Right nonDataInfodescr :: Either () (String,String))) $ keyfun grps)
      grpfun

groupAnalysis'_   :: DescriptiveKey key => ([a] -> (String, String)) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis'_ keydescrfun grpfun =
   groupAnalysis''_
      (map keydescrfun)
      grpfun

groupAnalysis_    :: DescriptiveKey key => (a -> (String, String)) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis_ keydescr grpfun = groupAnalysis'_ (keydescr . head) grpfun

----
-- Select grp combinators
----

selectGrpAnalysis''_  ::  DescriptiveKey key => String -> (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis''_ descr =
   Statistics.DescriptiveAnalysis.KeyableAC.selectGrpAnalysis'' (\_ -> selectGrpKey descr)


selectGrpAnalysis'_  ::  DescriptiveKey key => (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis'_  =   selectGrpAnalysis''_ ""

selectGrpAnalysis      ::  DescriptiveKey key => String -> (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis      =   selectGrpAnalysis''_

{-
-- Variant selectGrpKey :: (DataInfo b) => Maybe b -> String -> a

selectGrpAnalysis'' :: (DataInfo a, DescriptiveKey key) => String -> (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis'' descr =
   Statistics.DescriptiveAnalysis.KeyableAC.selectGrpAnalysis'' (\val -> selectGrpKey descr)

selectGrpAnalysis' :: (DataInfo a, DescriptiveKey key) => (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis' = selectGrpAnalysis'' ""

-}

----
-- Aggregation Combinators, based on aggregation analysis in PrimAC
-- using aggregateKey avoids using the combineList function
----


aggregateAnalysis'' :: DescriptiveKey key => String -> ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis'' descr aggrefun = 
   Statistics.DescriptiveAnalysis.PrimAC.aggregateAnalysis''' (\inputval aggreval keys -> aggregateKey descr keys) aggrefun

aggregateAnalysis' :: DescriptiveKey key => ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis' aggrefun = aggregateAnalysis'' "" aggrefun

aggregateAnalysis :: DescriptiveKey key => String -> ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis = aggregateAnalysis''

aggregateGrpsAnalysis  ::  DescriptiveKey key => String -> ([a] -> b) -> AnalysisFK key a b
aggregateGrpsAnalysis = aggregateAnalysis''

{-
-- Examples
aggregateAnalysis' :: (DescriptiveKey key) => ([a] -> String) -> ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis' keydescr aggregatefun =
   Statistics.DescriptiveAnalysis.PrimAC.aggregateAnalysis'' (aggregateKey . keydescr) aggregatefun

overallAverage' :: (Real a) => AnalysisFK String a Float  -- Real is required
overallAverage' = aggregateAnalysis' (\values -> "Average over " ++ show values) mean

aggregateAnalysis :: (DescriptiveKey key) => String -> ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis keydescr aggregatefun =
   Statistics.DescriptiveAnalysis.PrimAC.aggregateAnalysis' (aggregateKey keydescr) aggregatefun

overallAverage :: (Real a) => AnalysisFK String a Float  -- Real is really necessary
overallAverage = aggregateAnalysis "Average" mean

testAggrAna = overallAverage <.> (basicAnalysis "no. of logs" length) <.> groupAnalysisPerId
-}

----
-- Tupling combinators
----

entupleAnas :: (DescriptiveKey key) => AnalysisFK key a b1 -> AnalysisFK key a b2 -> AnalysisFK key a (b1, b2)
entupleAnas = Statistics.DescriptiveAnalysis.KeyableAC.entupleAnasWith' (\(orikey, k1, k2) -> tupleKey orikey k1 k2)

(<&>) :: (DescriptiveKey key) => AnalysisFK key a b1 -> AnalysisFK key a b2 -> AnalysisFK key a (b1, b2)
(<&>) = entupleAnas


{- Not a seperate primitive in the current framework

mapAnalysis'' :: (Keyable keya, Keyable keyO) =>
                 (keya -> [a] -> [keyb] -> [[b]] -> keyO)
                 -> ([(keya, a)] -> [(keyb, b)])
                 -> AnalysisFK keyO [a] [b]
mapAnalysis'' keyfun ana =
   Statistics.DescriptiveAnalysis.KeyableAC.mapAnalysis' -- mapAnalysis with startkey to kickstart inner analysis and combineAna for combining consequent analysis
      (\anainkey value anaoutkeys outvalue -> keyfun anainkey value anaoutkeys outvalue)
      ana

mapAnalysis' :: (DescriptiveKey key) => String -> AnalysisFK key a b-> AnalysisFK key [a] [b]
mapAnalysis' mapdescr ana =
   mapAnalysis'' (\anainkey value anaoutkeys outvalue -> mappingkey mapdescr {- head -} anaoutkeys) ana

mapAnalysis :: (DescriptiveKey key) => AnalysisFK key a b -> AnalysisFK key [a] [b]
mapAnalysis ana = mapAnalysis' "" ana
-}
