module Statistics.DescriptiveAnalysis.KeyHistoryListImpl where

import Statistics.DescriptiveAnalysis.KeyableAC (Keyable, combine, startkey, AnalysisFK )
import Statistics.DescriptiveAnalysis.DescriptiveKeyAC
                                                              ( DescriptiveKey
                                                              , basicKey
                                                              , groupKey
                                                              , selectGrpKey
--                                                               , mappingkey
                                                              , tupleKey
                                                              , aggregateKey
--                                                               , filteringkey
                                                              )

import Statistics.Presentation
import Statistics.BasicTypes
import Statistics.SummaryMeasures(calcFractions)
import Statistics.Completion
import Statistics.Ordering

import Data.List hiding ((\\))
import qualified Data.List ((\\))
import Data.Maybe
import Data.Typeable
import Data.Char
import Auxilary -- (traceVal, debug_key, stringToFilePath, trim, fst3, snd3, showMaybeData)
import Grouping
import System.Random

{-

Implementation of structured history like datatype with encapsulated field for the grouping subject, with possibility to access the DataInfo instance of the grouping subject


----
-- Descriptive key datatype, including presentation functions
----

Supported presentation:
showAsTable1D :: (Show b) => [(KeyHistory, b)] -> MarkUpDoc
showAsTable2D :: (Show a, Eq a, Ord a, Show b) => [(KeyHistory, [(a,b)])] -> MarkUpDoc
showAsTableDynamic :: (Show b) => [(KeyHistory, b)] -> MarkUpDoc
renderBarPlot1D :: (Num b, DataInfo b) => (b -> String) -> FilePath -> [(KeyHistory, b)] -> IO MarkUpDoc
renderBarPlot2D :: (Ordinal a, Num b, DataInfo a, DataInfo b) => AxisInfo a -> FilePath -> 
renderBarChart :: (Num b, Show b) => FilePath -> [(KeyHistory, b)] -> IO (FilePath, String)
renderStackedBarChart :: (Show a, Show b, Ord a, Num b) =>  FilePath -> [(KeyHistory, [(a,b)])] -> IO (FilePath, String)
renderBarChartDynamic :: (Num b, Show b) => FilePath -> [(KeyHistory, b)] -> IO (FilePath, String)
renderRelativeStackedBarChart :: (Num b, Real b, Show b) => FilePath -> [(KeyHistory, b)] -> IO (FilePath, String)
renderBoxPlot :: (Show b, Num b, Ord b) => FilePath -> [(KeyHistory, [b])] -> IO (FilePath, String)
showListAsCSV :: (Show b) => [(KeyHistory, [b])] -> String (undocumented in thesis)

Todo:

Presentation options:
- direction of axis titles
- title on/off
- filename handling, using Auxilary.isOnlyDirectory
- table totals
- renderDescr doesnt work for relative (aggr / tuple keys)
-}

keydebug = Auxilary.debug_key

newtype KeyHistory = KHL { getKHL :: [KeyHistoryDescriptor]}
  deriving (Eq, Ord, Show)

type KeyHistoryList = KeyHistory

-- A GADT KeyDescriptors
data KeyHistoryDescriptor where
   BasicAnalysis        :: String -> KeyHistoryDescriptor
   AggregateAnalysis    :: String -> [KeyHistory]-> KeyHistoryDescriptor
   GroupingAnalysis     :: DataInfo a => Either a (String, String) -> KeyHistoryDescriptor
   SelectGrpAnalysis    :: String -> KeyHistoryDescriptor
   TuplingAnalysis      :: KeyHistory -> KeyHistory -> KeyHistory -> KeyHistoryDescriptor

-- Grouping variants
--    GroupingAnalysis  :: (Show a, Eq a, DataInfo a) => Either a (String, String) -> KeyHistoryDescriptor
--    GroupingAnalysis  :: (Show a, DataInfo a) => Maybe a -> String -> String -> KeyHistoryDescriptor
--    GroupingAnalysis2 :: (a, AxisDataInfo a) -> String -> String -> KeyHistoryDescriptor

-- since we use a GADT, we define the following classes:

instance Show KeyHistoryDescriptor where
   show kd = 
      case kd of
         (BasicAnalysis descr) -> "BasicAnalysis: " ++ descr
         (GroupingAnalysis grpsubj) -> 
            let title = "GroupingAnalysis: " 
            in 
               either   (\realval -> title ++ "<DataInfo " ++ (shortname realval) ++ ": " ++ show realval ++ ">")
                        (\(subj, value) -> title ++ "(" ++ subj ++ ": " ++ value  ++ ")")  grpsubj
         (SelectGrpAnalysis descr) -> "SelectGrpAnalysis: " ++ descr
         (AggregateAnalysis descr keys) -> "AggregateAnalysis: " ++ descr ++ " keys: " ++ show {-(take 5 -} keys 
         (TuplingAnalysis orikey key1 key2) -> "TuplingAnalysis: (" ++ show orikey ++ ", " ++ show key1 ++ ", " ++ show key2 ++ ")"
         otherwise -> "Other key"

instance Eq KeyHistoryDescriptor where
  (==) =  (\d1 d2 ->
   case (d1,d2) of 
     (BasicAnalysis descr1, BasicAnalysis descr2) ->  descr1 == descr2
     (GroupingAnalysis grpsubj, GroupingAnalysis grpsubj2) -> show grpsubj == show grpsubj2  -- textual compare
     (SelectGrpAnalysis descr, SelectGrpAnalysis descr2) -> descr == descr2
     (AggregateAnalysis descr keys, AggregateAnalysis descr2 keys2) -> (descr == descr2) && (keys == keys2)
     (TuplingAnalysis orikey key1 key2, TuplingAnalysis orikeyB key1B key2B) -> (orikey == orikeyB) && (key1 == key1B) && (key2 == key2B)
     otherwise -> False -- problem with GHC GADT implementation error: "Error pattern not found in (==)"case
     )

-- Ord instance to be able to 'sort' the Descriptor, for grouping, when pretty printing
instance Ord KeyHistoryDescriptor where
   (<=) d1 d2 =  -- compare def. could be faster in use
      case (d1, d2) of
         (BasicAnalysis _, _) -> True
         (GroupingAnalysis _, BasicAnalysis _) -> False
         (GroupingAnalysis e1, GroupingAnalysis e2) -> -- comparing 2 GADT values is not possible: two rigid variables, so (compare val1 val2) == LT when Ord is part of DataInfo class, doesnt work. Fix Enum, fromEnum, returns Int, is comperable, but does all data /types fall in this class?
               if (either (shortname) (fst) e1) == (either (shortname) (fst) e2) then
                  case (e1,e2) of
                     (Left val1, Left val2) -> {- traceVal (show $ "val1: " ++ (show $ fromEnum val1) ++ "val2: " ++ (show $ fromEnum val2)) $ -} ordIndex val1 <= ordIndex val2
                           -- fromEnum val1 <= fromEnum val2 -- (compare val1 val2) == LT-- {- tmp test -}
                     (Right (_, str1), Right (_, str2)) -> str1 <= str2
                     otherwise -> error $ "no case for group compare " ++ show e1 ++ " || " ++ show e2
                     -- either (id) (snd) e1) <= (either (id) (snd) e2)
               else
                  (either (shortname) (fst) e1) <= (either (shortname) (fst) e2)
         (GroupingAnalysis _, _) -> True
         (SelectGrpAnalysis _, BasicAnalysis _) -> False
         (SelectGrpAnalysis _, GroupingAnalysis _) -> False
         (SelectGrpAnalysis _, _) -> True
         (AggregateAnalysis _ _, BasicAnalysis _) -> False
         (AggregateAnalysis _ _, GroupingAnalysis _) -> False
         (AggregateAnalysis _ _, SelectGrpAnalysis _) -> False
         (AggregateAnalysis _ _, _) -> True
         (TuplingAnalysis _ _ _, BasicAnalysis _) -> False
         (TuplingAnalysis _ _ _, GroupingAnalysis _) -> False
         (TuplingAnalysis _ _ _, SelectGrpAnalysis _) -> False
         (TuplingAnalysis _ _ _, AggregateAnalysis _ _) -> False
         (TuplingAnalysis t1a t1b t1c, TuplingAnalysis t2a t2b t2c) ->
            (t1b <= t2b)

instance Keyable KeyHistory where
   startkey = KHL []
   (KHL k1) `combine` (KHL k2) = KHL (k1 ++ k2)

instance DescriptiveKey KeyHistory where
   basicKey descr             = KHL [BasicAnalysis descr]
   groupKey grp               = KHL [GroupingAnalysis grp]
   selectGrpKey str           = KHL [SelectGrpAnalysis str]
   aggregateKey str keys      = KHL [AggregateAnalysis str keys]
   tupleKey orikey key1 key2  = KHL [TuplingAnalysis orikey key1 key2]

isBasicKeyD :: KeyHistoryDescriptor -> Bool
isBasicKeyD (BasicAnalysis _) = True
isBasicKeyD _               = False

isGrpKeyD (GroupingAnalysis _) = True
isGrpKeyD _               = False

isGrpSelectKeyD (SelectGrpAnalysis _) = True
isGrpSelectKeyD _               = False

isGrpAggrKeyD (AggregateAnalysis _ _) = True
isGrpAggrKeyD _ = False

isTuplingKeyD (TuplingAnalysis _ _ _) = True
isTuplingKeyD _ = False

-- filter on list of keydescriptors, not in structure of key
filterKeyOnBasics       :: KeyHistory -> KeyHistory
filterKeyOnBasics       = KHL . (filter isBasicKeyD) .  getKHL
filterKeyOnGrpings      = KHL . (filter isGrpKeyD) .  getKHL
filterKeyOnGrpSelects   = KHL . (filter isGrpSelectKeyD) .  getKHL
filterKeyOnGrpAggrs     = KHL . (filter isGrpAggrKeyD) .  getKHL
filterKeyOnGrpTuples    = KHL . (filter isTuplingKeyD) .  getKHL

filterOnlyGroupingsInKeyStructure :: KeyHistory -> KeyHistory   --  [Either (String,String) a]
filterOnlyGroupingsInKeyStructure (KHL descs) = KHL $ concatMap f descs
   where
      f (BasicAnalysis str) = []
      f (GroupingAnalysis eitherVal) = [(GroupingAnalysis eitherVal)]
      f (SelectGrpAnalysis str) = []
      f (AggregateAnalysis str keys) = concat $ map (getKHL . filterOnlyGroupingsInKeyStructure) keys
               -- getKHL $ filterOnlyGroupingsInKeyStructure $ foldl1 keysimilarity $ {- take 2-} keys
      f (TuplingAnalysis orikey k1 k2) = (getKHL $ filterOnlyGroupingsInKeyStructure orikey)
                -- ++ (getKHL $ filterOnlyGroupings k2) ++ (getKHL $ filterOnlyGroupings k1)
                -- getKHL $ filterOnlyGroupings $ keysimilarity k1 k2

buildListBy :: (a -> a -> [a]) -> [a] -> [a] -> [a]
buildListBy build l1 l2 = concat $ map (uncurry build) $ zip l1 l2

-- keysimilarity  only works in a foldl1 on a list of 2 values if the values are tuples or aggregates, these are flattened so further comparison to other tuples or aggregates fail, arizing all archieved data
keysimilarity :: KeyHistory -> KeyHistory -> KeyHistory
keysimilarity k1 k2 = KHL $ buildListBy similarkeydescriptor (getKHL k1) (getKHL k2)
      
-- descriptor similarty functions, returning an empty list if both descriptors are different or a singletong list of the descriptor with the descriptor, with recursive similar keys
-- this is because often you would like to maintain the structure of the keyhistory
similarkeydescriptor :: KeyHistoryDescriptor -> KeyHistoryDescriptor -> [KeyHistoryDescriptor]
similarkeydescriptor d1 d2 = 
          case (d1, d2) of
           (AggregateAnalysis descr keys, AggregateAnalysis descr2 keys2) ->
               if (descr == descr2) then  [AggregateAnalysis descr $ map (\(k1, k2) -> keysimilarity k1 k2) $ zip keys keys2] 
                                    else [] 
           (TuplingAnalysis orikey key1 key2, TuplingAnalysis orikeyB key1B key2B) ->
               [TuplingAnalysis (keysimilarity orikey orikeyB) (keysimilarity key1 key1B) (keysimilarity key2 key2B)]   
           (GroupingAnalysis e1, GroupingAnalysis e2) ->
               -- slightly more lose comparison of either values, based on the shortname / fst of the grouping subject
               if either (shortname) (fst) e1 == either (shortname) (fst) e2
               then [GroupingAnalysis e1]
               else []
               {-
               -- other more strict comparison of grouping subject
               case (eitherval1, eitherval2) of
                  (Left val1, Left val2) -> 
                      if (shortname val1) == (shortname val2)
                      then [GroupingAnalysis eitherval1] 
                      else []
                  (Right (strA1, strA2), Right (strB1, _)) ->
                      if (strA1 == strB1)
                      then [GroupingAnalysis (Right (strA1, strA2) :: Either () (String, String))]
                      else []
                  otherwise -> []  -}
           (desc1, desc2) -> if (desc1 == desc2) then [desc1] else []

-- list difference of keyhistory
-- Mind: (Data.List.\\) works for the simple case, based on equality of KeyDescriptors, example:
-- List> (List.\\) (getKHL key1) ([fromJust $ getLastGroupings key1])
-- [GroupingAnalysis: (Username: Username "peter"),BasicAnalysis: group size]
(\\) :: KeyHistory -> KeyHistory -> KeyHistory
k1 \\ k2 = KHL $ concat $ map (uncurry diffkeydescriptor) $ zip (getKHL k1) (tails $ getKHL k2)

-- return keyhistory diff, compared on first elm of snd param
diffkeydescriptor :: KeyHistoryDescriptor -> [KeyHistoryDescriptor] -> [KeyHistoryDescriptor]
diffkeydescriptor d1 d2s = 
          case d1 of
           (TuplingAnalysis orikey key1 key2) -> let diffkey = KHL d2s 
                                                 in  [TuplingAnalysis (orikey \\ diffkey) (key1 \\ diffkey) (key2 \\ diffkey)]
           otherwise -> 
              case (null d2s) of
               True -> [d1]
               False -> case (d1, head d2s) of  
                 (AggregateAnalysis descr keys, AggregateAnalysis descr2 keys2) ->
                     if (descr == descr2) 
                                   then [] 
                                   else [AggregateAnalysis descr $ map (\(k1, k2) -> keysimilarity k1 k2) $ zip keys keys2] 
                 (GroupingAnalysis eitherval1, GroupingAnalysis eitherval2) -> 
                     case (eitherval1, eitherval2) of
                       (Left val1, Left val2) -> 
                          if (shortname val1) == (shortname val2)
                          then []
                          else [GroupingAnalysis eitherval1] 
                       (Right (strA1, _), Right (strB1, _)) ->
                          if (strA1 == strB1)
                          then []
                          else [GroupingAnalysis (Right (strA1, "SimilarValue") :: Either () (String, String))]
                       otherwise -> []
                 (desc1, desc2) -> if (desc1 == desc2) then [] else [desc2]

type KHLAlgebra a =
   (  String -> a,
      ({-Maybe b -> -}(String, String) -> a),
      String -> a,
      String -> [[a]] -> a,
      [a] -> [a] -> [a] -> a
   )

-- fold only possible without use of the GADT feature, otherwise the type has to be known
foldKH = foldKHL
foldKHL :: {-(Show b) => -}KHLAlgebra a -> KeyHistory -> [a]
foldKHL alg@(basicAna, grpAna, selectAna, aggrAna, tupAna) (KHL descr) = map f descr
   where
      -- f :: KeyHistoryDescriptor -> a
      f (BasicAnalysis str) = basicAna str
      f (GroupingAnalysis grpeither) = either (\val -> grpAna (shortname val, head $ showContents val)) (grpAna) grpeither
      f (SelectGrpAnalysis str) = selectAna str
      f (AggregateAnalysis str keys) = aggrAna str $ map (foldKHL alg) keys
      f (TuplingAnalysis orikey key1 key2) = tupAna (foldKHL alg orikey)  (foldKHL alg key1) (foldKHL alg key2) 

foldKHL1 :: KHLAlgebra KeyHistoryDescriptor   -> KeyHistory -> KeyHistory
foldKHL1 alg kh = KHL $          foldKHL alg kh

foldKHL2 :: KHLAlgebra [KeyHistoryDescriptor] -> KeyHistory -> KeyHistory
foldKHL2 alg kh = KHL $ concat $ foldKHL alg kh

-- functions which use the (DataInfo b => b) field

-- pp key descr, showing the grouping value
ppKeySimple :: KeyHistory -> [String]
ppKeySimple (KHL descs) = map f descs -- foldKHL undefined
   where
      f (BasicAnalysis str) = str
      f (GroupingAnalysis eitherval) =  either (show) (\(str1,str2) -> str2) eitherval
      f (SelectGrpAnalysis str) = str
      f (AggregateAnalysis str _) = str
  --    f (TuplingAnalysis k1 k2) = "(" ++ ppKeySimple k1 ++ ", " ++ ppKeySimple k2 ++ ")"

ppKeyFull :: KeyHistory -> [String]
ppKeyFull (KHL descs) = map f descs -- foldKHL undefined
   where
      f (BasicAnalysis str) = str
--      f (GroupingAnalysis dv str1 str2) = str1 ++ ": " ++ str2
      f (GroupingAnalysis eitherval) = either (show) (\(str1,str2) -> str1 ++ ": " ++ str2) eitherval
      f (SelectGrpAnalysis str) = str
      f (AggregateAnalysis str _) = str
      f (TuplingAnalysis orikey key1 key2) = "tuple, unimpl"

ppKeyGeneral (KHL descs) = map f descs -- foldKHL undefined
   where
      f (BasicAnalysis str) = str
--      f (GroupingAnalysis dv str1 str2) = str1
      f (GroupingAnalysis eitherval) = either (shortname) (\(str1,str2) -> str1) eitherval
      f (SelectGrpAnalysis str) = str
      f (AggregateAnalysis str _) = str

-- returns the last applied basic or aggregation analysis, the tupled values are treated seperately
-- usefull for deciding on the columnheaders
shortTitles :: KeyHistory -> [String] -> [String]
shortTitles (KHL descs) startenv = foldl (.) id (map f descs) startenv
   -- concatMap (\descr -> f descr startenv) $ reverse descs -- foldKHL undefined
   where
--      f :: KeyDescriptor -> ([String] -> [String])
      f (BasicAnalysis str) = (\env -> if null env then [str] else env)
 --     f (GroupingAnalysis dv str1 str2) = id
      f (GroupingAnalysis eitherval) = id -- either (show) (\(str1,str2) -> str1 ++ ": " ++ str2) eitherval
      f (SelectGrpAnalysis str) = id
      f (AggregateAnalysis str _) = (\env -> if null env then [str] else env)
      f (TuplingAnalysis orikey key1 key2) = (\env -> env ++ (shortTitles key1 []) ++ (shortTitles key2 []))

----
-- Special Visualisation Functions for KeyHistory key
----

-- Helper functions, to Auxilary
capitalize str =
   if length strippedstr > 0
   then (toUpper $ head strippedstr) : tail strippedstr
   else strippedstr
   where strippedstr = trim str

prepend :: String -> [String] -> [String]
prepend pre = 
   foldl (\sels val -> sels ++ [pre] ++ [val]) [] 

postpend :: String -> [String] -> [String]
postpend post =
   foldr (\str rest -> rest ++ [post] ++ [str]) []

-- remove comments between brackets
ignore_comments str =
   takeWhile (not . bracketOpen) str -- use parsercomb
   where
      bracketClose c = c ==')'
      bracketOpen c = c =='('

-- v1 title
titleold :: KeyHistory -> String
titleold key = 
   {-columnsdescr++" "++-}aggrs++bass++grps
   where
     bass  = capitalize  $ concat $ intersperse " of " $ reverse $ filter (not . (=="")) $ ppKeySimple $ filterKeyOnBasics key
     grps  = prepend' " per " $ ppKeyGeneral $ filterOnlyGroupingsInKeyStructure key  -- filterOnlyGroupingsInKeyStructure has been changed, now returning, all groupings in all subkeys
     aggrs = capitalize  $ concat $ intersperse " of " $ reverse $ filter (not . (=="")) $ ppKeySimple $ filterKeyOnGrpAggrs key
     prepend' pre strs = foldr (\str rest -> rest ++ "" ++ pre ++ "" ++str) "" strs

basicAStr   = "" -- "of"
tupleAStr   = "and"
groupAStr   = "per"
aggrAStr    = "of"
statGrp     = "for"

-- v2 title
fullTitle :: KeyHistory -> [String] -> [String]
fullTitle (KHL descs) startenv = foldr (f) startenv $ reverse descs
   where
--    f :: KeyDescriptor -> ([String] -> [String])
      f (BasicAnalysis str) = (\env -> if null str then {- []: -}env else basicAStr:str:env)
      f (GroupingAnalysis eitherval) = 
            (\env -> groupAStr:(either (shortname) (fst) eitherval):env)
      f (SelectGrpAnalysis str) = id
      f (AggregateAnalysis str keys) = (\env ->
            let keysim = foldl1 keysimilarity keys
            in
               if null str
               then fullTitle keysim [] ++ env -- map (//keysim) keys
               else aggrAStr:str:(fullTitle (uniqkeygrps $ filterOnlyGroupingsInKeyStructure keysim) []) ++ env)

      f (TuplingAnalysis orikey key1 key2) = (\env ->
            let
               fsttuple_ana = fullTitle (key1\\orikey) []
               sndtuple_ana = fullTitle (key2\\orikey) []
               -- only show the tuplestr between two actual values
               andstr = if null fsttuple_ana || null sndtuple_ana then [] else [tupleAStr]  
            in
               fsttuple_ana ++ andstr ++
               sndtuple_ana ++ fullTitle orikey [] ++ env)

-- grps the internal keycdescriptors and filters them, returning of each analysis kind one
uniqkeygrps :: KeyHistory -> KeyHistory
uniqkeygrps key =
   KHL $ concat $ filter ((1==) . length) $ map nub $ {- groupAll -} groupAllByEq (\d1 d2 -> (length $ similarkeydescriptor d1 d2) > 0) $ sort $ getKHL $ key

combinekeys (KHL ds1) (KHL ds2) = KHL (ds1 ++ ds2)

-- should also work for flattened analysis
rowTitle = ppGrpsOnly
ppGrpsOnly :: KeyHistory -> String
ppGrpsOnly key = 
         let
            -- grouping on first level of keyhistory
            fstlevelgrps = filterKeyOnGrpings key
            -- groupings in aggregation, tuples etc.
            uniqgrps = uniqkeygrps $ filterOnlyGroupingsInKeyStructure key
         in  if ((length $ getKHL fstlevelgrps) > 0 || (length $ getKHL uniqgrps) > 0)
             then concat $ intersperse " / " $ ppKeySimple $ combinekeys fstlevelgrps (plaindiff uniqgrps fstlevelgrps) -- (uniqgrps\\fstlevelgrps)  -- the latter shows 'SimilarValue'
             else "Complete data set" -- ++ show key

-- table column header or axis title (single value)
columns :: DataInfo a => KeyHistory -> a -> [String]
columns key value = 
   if (length $ getKHL key) > 0  -- check from barplot1d
   then
      take (length $ showContents value) $ shortTitles key [] ++ repeat "data"  -- showing all tupled analysis, a optional preapplied analysis might ruin this feature
   else
      repeat "data"

-- v3 title, shows the basicAnas, then groupings, then invariant group values
renderDescr :: [KeyHistory] -> [String]
renderDescr keys =
   let fixedgrpelms = filterKeyOnGrpings $ getSameKeyDescrList keys
         -- filterOnlyGroupingsInKeyStructure $ getSameKeyDescrList keys
       otherelms = plaindiff (foldl1 keysimilarity keys) fixedgrpelms
   in fullTitle (otherelms) [] ++
         (prepend statGrp (ppKeyFull fixedgrpelms))
   where

      -- returns a key with the exact same descriptors, sim. to keysimilarity
      getSameKeyDescr :: KeyHistory -> KeyHistory -> KeyHistory
      getSameKeyDescr k1 k2 = KHL $ map fst $ filter ((uncurry (==))) $ zip (getKHL k1) (getKHL k2)
            -- above doesnt work for aggre/ tupled keys, rewrite like in keysimilarity , bus now for stricter case

      -- get constant descr into new key, by comparing subsequent logs
      getSameKeyDescrList :: [KeyHistory] -> KeyHistory
      getSameKeyDescrList keys = maybe (KHL []) id (foldl (cmpsimilarkeys) Nothing keys)
         where
            cmpsimilarkeys :: Maybe KeyHistory -> KeyHistory -> Maybe KeyHistory
            cmpsimilarkeys sim_sofar k =
               case sim_sofar of
                  Nothing -> Just k
                  Just k0 -> Just $ getSameKeyDescr k0 k

plaindiff :: KeyHistory -> KeyHistory -> KeyHistory
plaindiff (KHL kd1) (KHL kd2) = KHL $ foldl addIfNotKd2 [] kd1
   where
      addIfNotKd2 list d =
         if any (==d) kd2 then (list++[]) else (list++[d])

----
-- Ungrouping: [(KeyHistory, b)] -> Maybe [(KeyHistory, [(String, b)])]
----

-- move the last grouped value into the value part of the analysis result
-- mind completion and mind ordering, should 
ungroup1D :: [(KeyHistory, b)] -> Maybe [(KeyHistory, [(String, Maybe b)])]
ungroup1D [] = Nothing
ungroup1D input@(x:xs) = 
   let
      grpsubj = getLastGroupings $ fst x -- last grp type, to be moved to value field
      -- ungroupLastGroup :: [(KeyHistory, b)] -> KeyHistoryDescriptor -> [(KeyHistory, [(String, Maybe b)])]
      ungroupLastGroup inp grpdescr =
         let
            groupOnAllGrpsButLastGrp :: [(KeyHistory,b)] -> [[(KeyHistory,b)]]
            groupOnAllGrpsButLastGrp =
               groupAllUnder
                  (\elm -> (getKHL $ fst elm) Data.List.\\ [getLastGroupingsExp $ fst elm])
            newkey :: KeyHistory  -> KeyHistory
            newkey oldkey = KHL $ (getKHL oldkey) Data.List.\\ [getLastGroupingsExp oldkey]
            extractInfo :: (KeyHistory, b) -> (KeyHistoryDescriptor, b) 
            extractInfo (key, value) =
                     (-- KHL $ (getKHL key) Data.List.\\ [getLastGroupingsExp key],
                     getLastGroupingsExp key,   -- keydescriptor
                     value)
            -- extractGrp ::  [KeyHistoryDescriptor] -> [(KeyHistory, (a, b))] -> (KeyHistory, [(String, Maybe b)])
            extractGrp allvalues elms =
               let   -- complementedAndOrderData :: {-[KeyHistoryDescriptor] -> [(k, v)] ->-} [(KeyHistoryDescriptor, Maybe b)]
                  complementedAndOrderData =
                     Data.List.sortBy (\elm1 elm2 -> compare (fst elm1) (fst elm2))  -- sort on grpDescription
                     $ (complementVals allvalues)  -- complement based on internally available values
                     $ map extractInfo elms
               in
                  (newkey $ fst $ head elms,
                     map (fmapTupleFst (\(GroupingAnalysis esubj) -> either show snd esubj))
                         complementedAndOrderData)
            allInternalLastGrps allgrps = nub $ map (fst . extractInfo) allgrps
            -- complementVals :: (Eq a) => [a] -> [(a, (a1, b))] -> [(a, Maybe b)]
            complementVals allvals = (complementGen allvals Nothing) . map (\(k,v) -> (k, Just v))
         in map 
               (extractGrp (allInternalLastGrps inp)) (groupOnAllGrpsButLastGrp inp)
   in
      maybe Nothing (Just . ungroupLastGroup input) grpsubj

getLastGroupings :: KeyHistory -> Maybe KeyHistoryDescriptor
getLastGroupings = headMaybe . reverse . getKHL . filterOnlyGroupingsInKeyStructure

-- if key is garanteed to have a last grouped elm
getLastGroupingsExp :: KeyHistory -> KeyHistoryDescriptor
getLastGroupingsExp key =
   maybe (malformedkeyError "ungrouping") id (getLastGroupings key)

malformedkeyError :: String -> a
malformedkeyError subj = error $ "Error: Malformed keys while " ++ subj

headMaybe :: [a] -> Maybe a
headMaybe [] = Nothing
headMaybe (x:xs) = Just x

longestListFromLists :: [[a]] -> [a]
longestListFromLists lists = head $ filter ((== maxGrpLen) . length) lists
            where maxGrpLen = maximum $ map length lists

----
-- Presentation rendering functions providing a simpler approach to the generic presenations
----

-- Todo: multiple bewerking samenvoegen in column header, (bijv. afronden na berek van aantal)
showAsTable1D :: (Show b) => [(KeyHistory, b)] -> MarkUpDoc
showAsTable1D [] =
   markUpTable1D Nothing ("No data to present in table") "data" ([]::[(String,[String])])
showAsTable1D tabdata@((key,_):_) =
   let
      keyelms = map fst tabdata
      keyelm = head keyelms
      tabdataelms = map snd tabdata
      tabdataelm = head tabdataelms
      descr key = capitalize  $ unwords $ renderDescr keyelms -- fullTitle key []
 --      calcs = aggrs ++ bass
--      columnsdescr = concat $ intersperse " and " columns
      tabdataToShow = map (\(v1,v2) -> (rowTitle v1, show v2)) $ sortByFstElm tabdata
   in
      markUpTable1D 
         Nothing 
         (descr keyelm ++ if keydebug then " orikey (head): " ++ show keyelm else "") 
         (capitalize  $ concat $ intersperse " and " (shortTitles key []))
         tabdataToShow

showAsTable2D :: (Show a, Eq a, Ord a, Show b) => [(KeyHistory, [(a,b)])] -> MarkUpDoc
showAsTable2D [] =
   markUpTable1D Nothing ("No data to present in table") "data" ([]::[(String,[String])])
showAsTable2D tabdata@((key,_):_) = 
         let
            keyelms = map fst tabdata
            descr = capitalize  $ unwords $ renderDescr keyelms
            allvalues2dDim = nub $ concatMap ((map fst) . snd) tabdata
            cols = map show $ sort allvalues2dDim
            complementData allvals = map (fmapTupleSnd $ complementGen' allvals)
            newtabdata = map (fmapTuple (rowTitle, (map (showMaybeData . snd))))
                       $ sortByFstElm                       -- sort the 'outer' dimension(s)
                       $ map (fmapTupleSnd sortByFstElm)    -- sort the 'inner' dimension
                       $ complementData allvalues2dDim      -- complement in 2nd dimension
                           tabdata
         in
            markUpTable2D 
               Nothing 
               (descr ++ if keydebug then " orikey (head): " ++ show key else "") 
               cols
               newtabdata

showAsTableDynamic :: (Show b) => [(KeyHistory, b)] -> MarkUpDoc
showAsTableDynamic [] =
   markUpTable1D Nothing ("No data to present in table") "data" ([]::[(String,[String])])
showAsTableDynamic tabdata@((key,_):_) = 
   case ungroup1D tabdata of
      Nothing          -> showAsTable1D tabdata
      Just newgrouping -> 
         let
            keyelms = map fst tabdata
            descr = capitalize  $ unwords $ renderDescr keyelms

            -- we assume completion and ordering on the new dimension
            -- completions can not be correctly handled by a seperated form of 'internal' completion, since we then lose the ordering
            cols = map fst $ snd $ head newgrouping
            newtabdata = map (\(v1, list) -> (rowTitle v1, (map (showMaybeData . snd) list))) newgrouping
         in
            markUpTable2D 
               Nothing 
               (descr ++ if keydebug then " orikey (head): " ++ show key else "") 
               cols
               newtabdata

-- special DI feature to get 2d elements, by using showContents, sort of variant of showAsTable_Dynamic
-- but this solution requires every presented value to have DI instance, also tupled values, lists and so on
showAsMarkUpTable1D :: (Show b, DataInfo b) => [(KeyHistory, b)] -> MarkUpDoc
showAsMarkUpTable1D = showAsTable'

showAsTable' :: (Show b, DataInfo b) => [(KeyHistory, b)] -> MarkUpDoc
showAsTable' [] =
   markUpTable1D Nothing ("No data to present in table") "data" ([]::[(String,[String])])
showAsTable' tabdata@((key,_):_) = 
   let
      keyelms = map fst tabdata
      tabdataelms = map snd tabdata
      tabdataelm = head tabdataelms
      descr = capitalize  $ unwords $ renderDescr keyelms
 --      calcs = aggrs ++ bass
--      columnsdescr = concat $ intersperse " and " columns
      tabdataToShow = map (\(v1,v2) -> (rowTitle v1, showContents v2)) tabdata
   in
      markUpTable2D 
         Nothing 
         (descr ++ if keydebug then " orikey (head): " ++ show key else "") 
         (columns key tabdataelm) {-[ignore_comments calcs]-} 
         tabdataToShow

{-
Plots are rendered by an external tool, making them IO Monad functions

-- important:
The plot (or ploticus, the rendering tool) prefabs first requires a data file to be stored somewhere
-}

-- Render functions with completion and ordering on the actual grouping subject
-- old style
renderBarPlot1D :: (Num b, DataInfo b) => (b -> String) -> FilePath -> [(KeyHistory, b)] -> IO MarkUpDoc
renderBarPlot1D yshow fp plotdata =
   let desc = capitalize  $ unwords $ renderDescr (map fst plotdata)
       name = capitalize  $ unwords $ shortTitles key1 []
       key1 = head $ map fst plotdata
       val1 = head $ map snd plotdata
   -- render plot, MIND: do not change the name is relevant for the reference to the data file
   in
      if length plotdata > 0
      then
         do 
         renderPNGPlot fp (stringToFilePath desc) $
               barPlotWithComplAndOrder
                  (Just (AxisInfo (Right (compare :: KeyHistory -> KeyHistory -> Ordering)) rowTitle))
                  (Just $ AxisInfo (Right (error "no compare for num value in barplot")) yshow)
                  desc (columns key1 val1) plotdata
         return $ figure Nothing desc (fp, stringToFilePath desc ++ ".png")
     else
        return $ figure Nothing "No data" (fp, "")

renderBarPlot2D :: (Ordinal a, Num b, DataInfo a, DataInfo b) => AxisInfo a -> FilePath -> [(KeyHistory, [(a,b)])] -> IO MarkUpDoc
renderBarPlot2D yaxisinfo fp plotdata@((key1, val1):_) = do
   let name = capitalize  $ unwords $ shortTitles key1 [] -- ++  ["stacked barplot"]
       desc = unwords $ "Bar plot" : fullTitle key1 []  -- check string if correct filepath
   --render plot
   renderPNGPlot fp (stringToFilePath desc) $ 
      barPlot2DWithOptComplAndOrder (Just (AxisInfo (Right $ {-(\e1 e2 -> compare (fromEnum e1) (fromEnum e2))-}compare) rowTitle)) (Just yaxisinfo) name ("","") plotdata 
   return (
      if length plotdata > 0
      then figure Nothing desc (fp, stringToFilePath desc ++".png")
      else figure Nothing "No data" (fp,"")
      )

----
-- Bar charts as described in thesis
----

renderBarChart :: (Num b, Show b) => FilePath -> [(KeyHistory, b)] -> IO (FilePath, String)
renderBarChart fp []       = do 
   plotfile <- renderPNGPlot fp (stringToFilePath "No data") $
               barPlot "No data" ("","") ([]::[(String,Int)])
   return (plotfile, "Barplot no data")
renderBarChart fp plotdata@((key,val):_) =
   let descr = capitalize  $ unwords $ "Bar chart of " : (renderDescr {-$ causes a lot of title pbs!! map (KHL . sort . getKHL)-} $ (map fst plotdata))
       newplotdata = map (fmapTupleFst rowTitle)
                   $ sortByFstElm plotdata
       groups = ppGrpsOnly  $ foldl1 keysimilarity $ map fst plotdata
   in do 
         plotfile <- renderPNGPlot fp (stringToFilePath descr) $
            barPlot
               ({-traceVal (show $ map fst plotdata) $ -}descr)
               (groups, capitalize $ concat $ intersperse " and " (shortTitles key []))
               newplotdata
         return (plotfile, descr)

renderStackedBarChart :: (Show a, Show b, Ord a, Num b) =>  FilePath -> [(KeyHistory, [(a,b)])] -> IO (FilePath, String)
renderStackedBarChart fp []       = do 
   plotfile <- renderPNGPlot fp (stringToFilePath "No data") $
               barPlot "No data" ("","") ([]::[(String,Int)])
   return (plotfile, "Stacked bar chart no data")
renderStackedBarChart fp plotdata@((key, val):_) =
   let descr = capitalize  $ unwords $ "Stacked bar chart of" : renderDescr (map fst plotdata)
       allvalues2dDim = nub $ concatMap ((map fst) . snd) plotdata
       cats = map show $ sort allvalues2dDim
       complementData allvals = map (fmapTupleSnd $ complementGen' allvals)
       newplotdata = map (fmapTuple (rowTitle, (map (maybe 0 id . snd))))
                  $ sortByFstElm                       -- sort the 'outer' dimension(s)
                  $ map (fmapTupleSnd sortByFstElm)    -- sort the 'inner' dimension
                  $ complementData allvalues2dDim      -- complement in 2nd dimension
                       plotdata
       groups = ppGrpsOnly $ foldl1 keysimilarity $ map fst plotdata
   in do
         plotfile <- renderPNGPlot fp (stringToFilePath descr) $ 
            stackedBarPlot 
               (traceVal (show $ map fst plotdata) $ descr)
               (groups,"")
               cats
               newplotdata
               []
         return (plotfile, descr)

-- used to be renderBarPlotDynamic
renderBarChartDynamic :: (Num b, Show b) => FilePath -> [(KeyHistory, b)] -> IO (FilePath, String)
renderBarChartDynamic fp [] =
   do
      plotfile <- renderPNGPlot fp (stringToFilePath "No data") $
                     barPlot "No data" ("", "") ([]::[(String,Int)])
      return (plotfile, "Stacked bar chart  no data")
renderBarChartDynamic fp plotdata@((key, val):_) =
   case ungroup1D plotdata of
      Nothing          -> renderBarChart fp plotdata
      Just newgrouping -> 
         let
            keyelms = map fst plotdata
            descr = capitalize $ unwords $ "Stacked bar chart of" : renderDescr keyelms -- key 

            -- we assume completion and ordering on the new dimension
            -- completions can not be correctly handled by a seperated form of 'internal' completion, since we then lose the ordering
            cats = map fst $ snd $ head newgrouping
            newplotdata = map (fmapTuple (rowTitle, map (maybe 0 id . snd))) newgrouping
         in
            do
               plotfile <- renderPNGPlot fp (stringToFilePath descr) $ 
                              stackedBarPlot 
                                 descr
                                 ("","") 
                                 cats
                                 newplotdata
                                 []
               return (plotfile, descr)

-- used to be renderBarPlotDynamic
renderRelativeStackedBarChart :: (Num b, Real b, Show b) => FilePath -> [(KeyHistory, b)] -> IO (FilePath, String)
renderRelativeStackedBarChart fp [] =
   do
      plotfile <- renderPNGPlot fp (stringToFilePath "No data") $
                     barPlot "No data" ("", "") ([]::[(String,Int)])
      return (plotfile, "Relative bar chart no data")
renderRelativeStackedBarChart fp plotdata@((key, val):_) =
   case ungroup1D plotdata of
      Nothing          -> renderBarChart fp plotdata  -- make rel. version
      Just newgrouping -> 
         let
            keyelms = map fst plotdata
            descr = capitalize $ unwords $ "Relative bar chart of" : renderDescr keyelms -- key

            -- we assume completion and ordering on the new dimension
            -- completions can not be correctly handled by a seperated form of 'internal' completion, since we then lose the ordering
            cats = map fst $ snd $ head newgrouping  -- calcFractions
            newplotdata = map (fmapTuple (rowTitle, calcFractions . map (maybe 0 id . snd))) newgrouping
         in
            do
               plotfile <- renderPNGPlot fp (stringToFilePath descr) $ 
                              stackedBarPlot 
                                 descr
                                 ("","percentages (%)") 
                                 cats
                                 newplotdata
                                 [yrange 0 100, singlelnlegend]
               return (plotfile, descr)


-- Ord b req to deside on max
renderBoxPlot :: (Show b, Num b, Ord b) => FilePath -> [(KeyHistory, [b])] -> IO (FilePath, String)
renderBoxPlot fp [] = do
   plotpath <- renderPNGPlot fp (stringToFilePath "Empty box plot, no data") 
          (rangeBarPlot "No data" ("","") [])
   return (plotpath, "Empty box plot, no data")
renderBoxPlot fp plotdata@((key, val):_) = do
   let 
       name = capitalize  $ unwords $ shortTitles key []
       desc = unwords $ "Box plot" : fullTitle key []
       newplotdata = map (fmapTupleFst rowTitle)
                   $ sortByFstElm plotdata
   plotpath <- renderPNGPlot fp (stringToFilePath desc) $
                  rangeBarPlot 
                     desc 
                     ("","") 
                     newplotdata
   return (plotpath, desc)

renderRangeBarPlot :: (Num b, Ord b, Show b) => FilePath -> [(KeyHistory, [b])] -> IO (FilePath, String)
renderRangeBarPlot = renderBoxPlot
{-
formGNUPlotBar :: (DataInfo b) => [(KeyHistory,b)] -> FilePath -> String
formGNUPlotBar input outfile = undefined -- 

   let
      keyelms = map fst input
      keyelm = head keyelms
      constantInputFactor = foldl1 (getKeySimilarity) keyelms -- requires non empty list
      fixedGroupes = getGroupsKH constantInputFactor
      allGroups = getGroupsKH keyelm
      nonFixedGroups = listToKH $ allGroups \\ fixedGroupes

      title =
            if fixedGroupes == [] then getCalculatedValueDescr constantInputFactor ++ " " ++ getAllGroupDescrBasic nonFixedGroups
                                  else getCalculatedValueDescr constantInputFactor ++ " for " ++ getAllGroupDescr constantInputFactor ++ " " ++ getAllGroupDescrBasic nonFixedGroups
            ++ " Histogram"
      xaxistitle = getAllGroupDescrBasic nonFixedGroups
      yaxistitle = getCalculatedValueDescr keyelm
      -- gnudata = getGroupNameKeyTransfKeyHistory input
   in
      toGNUPlotBar title (xaxistitle, yaxistitle) outfile input
-}


{-
plotTool :: String -> FilePath -> IO ()
plotTool = gnutool

showPlotBar :: [(KeyHistory, b)] -> MarkUpDoc
showPlotBar tabdata = 
   let
      keyelms = map fst tabdata
      keyelm = head keyelms
      tabdataelms = map snd tabdata
      tabdataelm = head tabdataelms
--      colheader = if isJust $ description tabdataelm 
--                  then fromJust $ description tabdataelm 
--                  else shortname tabdataelm
      bass = capitalize  $ concat $ intersperse " of " $ reverse $ filter (not . (=="")) $ ppKeySimple $ filterKeyOnBasics keyelm
      grps  = prepend " per " $ ppKeyGeneral $ filterKeyOnGrpings keyelm
      aggrs = capitalize $ concat $ intersperse " of " $ reverse $ filter (not . (=="")) $ ppKeySimple $ filterKeyOnGrpAggrs keyelm
      
      calcs = aggrs ++ bass
      ppGrpsOnly key = concat $ intersperse " / " $ ppKeySimple $ filterKeyOnGrpings key
      columns = take (length $ showContents tabdataelm) $ shortTitles keyelm [] ++ repeat "data"  -- showing all tupled analysis, a optional preapplied analysis might ruin this feature
      tabdataToShow = map (\(v1,v2) -> (ppGrpsOnly v1, showContents v2)) tabdata
   in
      if length tabdata > 0 then
         figure Nothing (aggrs++bass++grps) (xaxistitle, yaxistitle) tabdataToShow
      else
         figure Nothing ("No data available") ("x", "y") tabdataToShow

showImage = showPlotBar
-}
showListAsCSV :: (Show b) => [(KeyHistory, [b])] -> String
showListAsCSV = renderAsCSVFile2D
