module Statistics.DescriptiveAnalysis.KeyableAC where

import qualified Statistics.DescriptiveAnalysis.PrimAC

{-
-- for combinator examples
import Data.List
import Data.Maybe

import Auxilary
import Statistics
import Logging
import Grouping.Logging
import Metrics
import Select
-}


type AnalysisFK key a b = Statistics.DescriptiveAnalysis.PrimAC.Analysis key a key b 

-- Monoid class like Keyable, fixes key type, limiting the ease of presentation lateron
class Keyable a where
   startkey    :: a  -- keystarter, empty key, reset key
   combine     :: a -> a -> a
   combineList :: [a] -> a -> a
   combineList keys = flip (foldl combine) keys

(<.>) = (Statistics.DescriptiveAnalysis.PrimAC.<.>)

-- runAnalysis combinator redefined
runAnalysis :: (Keyable key) => a -> AnalysisFK key a b -> [(key, b)]
runAnalysis input ana = Statistics.DescriptiveAnalysis.PrimAC.runAnalysis input startkey ana

peekKey :: (Keyable key) => AnalysisFK key a b -> [key]
peekKey = map fst . (runAnalysis $ error "Error occured, while peeking for key")

getAnaResult :: (Keyable key) => a -> AnalysisFK key a b -> [b]
getAnaResult input = map snd . runAnalysis input

-- basic analysis combinators, specialisations of the generic version in PrimAC.hs
basicAnalysis'' :: Keyable key => (a -> b -> key) -> (a -> b) -> AnalysisFK key a b
basicAnalysis'' keyfun fun =
   Statistics.DescriptiveAnalysis.PrimAC.basicAnalysis''' (\value outcome oldkey -> combine oldkey $ keyfun value outcome) fun

basicAnalysis' :: Keyable key => (a -> key) -> (a -> b) -> AnalysisFK key a b
basicAnalysis' keyfun fun = Statistics.DescriptiveAnalysis.PrimAC.basicAnalysis'' (\value oldkey -> combine oldkey $ keyfun value) fun

basicAnalysis_ :: Keyable key => key -> (a -> b) -> AnalysisFK key a b
basicAnalysis_ key = basicAnalysis' (\val -> key)
-- Statistics.DescriptiveAnalysis.PrimAC.basicAnalysis'' (\value oldkey -> combine oldkey $ keyfun value) fun

basicAnalysis :: Keyable key => key -> (a -> b) -> AnalysisFK key a b
basicAnalysis = basicAnalysis_

groupAnalysis''  :: Keyable key => ([[a]] -> [key]) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis'' keyfun grpfun =
   Statistics.DescriptiveAnalysis.PrimAC.groupAnalysis'' (\grps oldkey -> map (combine oldkey) $ keyfun grps) grpfun

groupAnalysis'   :: Keyable key => ([a] -> key) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis' keyfun grpfun = groupAnalysis'' (map keyfun) grpfun
-- Statistics.DescriptiveAnalysis.PrimAC.groupAnalysis' (\grps oldkey -> combine oldkey $ keyfun grps) grpfun

groupAnalysis_   :: Keyable key => (a -> key) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis_ keyfun grpfun = groupAnalysis' (\grp -> keyfun $ head grp) grpfun

groupAnalysis    ::  Keyable key => (a -> key) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]
groupAnalysis    =   groupAnalysis_ 

{-
-- easy grouping function if [a] -> [(b,[a])] is available, like in Grouping.Logging
-- this is not necc when using:
groupAnalysis'' :: (Keyable key) => ([[a]] -> [key]) -> ([a] -> [[a]]) -> AnalysisFK key [a] [a]

groupAnalysisSpec :: (Keyable key) => (b -> key) -> ([a] -> [(b, [a])]) -> AnalysisFK key [a] [a]
groupAnalysisSpec keyfun grpfun =
     concatMap newAna
   where
      newAna (orikey, list) = map (\(grpId, selection) -> (combine orikey (keyfun grpId), selection)) $ grpfun list
         -- enTupleGroup :: (a -> b) -> [[a]] -> [(b,[a])], usefull here?
-}

aggregateAnalysis'' :: Keyable key => ([a] -> b -> key) -> ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis'' keyfun aggregatefun =
   Statistics.DescriptiveAnalysis.PrimAC.aggregateAnalysis''' (\values aggre keys -> combineList keys (keyfun values aggre)) aggregatefun 

aggregateAnalysis' :: Keyable key => ([a] -> key) -> ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis' keyfun aggregatefun =
   aggregateAnalysis'' (\values _ -> keyfun values) aggregatefun

aggregateAnalysis :: Keyable key => key -> ([a] -> b) -> AnalysisFK key a b
aggregateAnalysis key aggregatefun = aggregateAnalysis' (\_ {- values -} -> key) aggregatefun

selectGrpAnalysis''  ::  Keyable key => (a -> key) -> (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis'' keyfun predfun = 
   Statistics.DescriptiveAnalysis.PrimAC.selectGrpAnalysis'' (\value orikey -> combine orikey (keyfun value)) predfun

selectGrpAnalysis'   ::  Keyable key => key -> (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis' key = selectGrpAnalysis'' (\_ -> key)

selectGrpAnalysis    ::  Keyable key => key -> (a -> Bool) -> AnalysisFK key a a
selectGrpAnalysis    =   selectGrpAnalysis'

-- previous implementation
filterAna :: Keyable key => key -> (AnalysisFK key a Bool) -> AnalysisFK key [a] [a]
filterAna key selectana =
   Statistics.DescriptiveAnalysis.PrimAC.filterAna' (\oldkey -> combine oldkey key) selectana

-- List manipulation functions, filter, map, etc
-- filterAna' :: (Keyable key) => [a] -> key -> (AnalysisFK key a Bool) -> AnalysisFK key [a] [a]
-- filterAna' keyfun selectana =
--    Statistics.DescriptiveAnalysis.PrimAC.filterAna'' (\values oldkey -> combine oldkey $ keyfun values) selectana


-- tupling based on the original, and the two new keys
entupleAnasWith' :: ((key, key, key) -> key) -> AnalysisFK key a b1 -> AnalysisFK key a b2 -> AnalysisFK key a (b1, b2)
entupleAnasWith' keyfun ana1 ana2 = (\input ->
   let ana1result = ana1 input 
       ana2result = ana2 input
       orikeys = map fst input
   in 
      zip (map (keyfun) $ zip3 (map fst input) (map fst ana1result) (map fst ana2result)) (zip (map snd ana1result) (map snd ana2result))
   )

-- tupling based on only the two new keys
entupleAnasWith :: ((key, key) -> key) -> AnalysisFK key a b1 -> AnalysisFK key a b2 -> AnalysisFK key a (b1, b2)
entupleAnasWith keyfun ana1 ana2 = (\input ->
   let ana1result = ana1 input 
       ana2result = ana2 input
   in 
      zip (map (keyfun) $ zip (map fst ana1result) (map fst ana2result)) (zip (map snd ana1result) (map snd ana2result))
   )

entupleAnas :: (Keyable key) => AnalysisFK key a b1 -> AnalysisFK key a b2 -> AnalysisFK key a (b1, b2)
entupleAnas = entupleAnasWith (\(k1, k2) -> combine k1 k2)

{-
-- mapAnalysis'' :: (Keyable keyI, Keyable keyO) => (keyI -> [a] -> [keyI] -> [[b]] -> keyO -> keyO) -> AnalysisFK keyI a b -> AnalysisFK keyO [a] [b]
-- mapAnalysis'' :: (Keyable key) => (key -> [a] -> [key] -> [[b]] -> key -> key) -> AnalysisFK key a b -> AnalysisFK key [a] [b]
mapAnalysis'' :: Keyable keya => (keya -> [a] -> [keyb] -> [[b]] -> keyc -> keyd)
                                 -> ([(keya, a)] -> [(keyb, b)])
                                 -> Statistics.DescriptiveAnalysis.PrimAC.Analysis keyc [a] keyd [b]
-- possible since only empty is used and not yet combine, which would require an fixed key for the resulting analysis
mapAnalysis'' keyfun ana =
   Statistics.DescriptiveAnalysis.PrimAC.mapAnalysis'''
      startkey
      keyfun
      ana

mapAnalysis' :: (Keyable keyc, Keyable keya) =>
                (keya -> [a] -> [keyb] -> [[b]] -> keyc)
                -> ([(keya, a)] -> [(keyb, b)])
                -> AnalysisFK keyc [a] [b]
mapAnalysis' keyfun ana =
   mapAnalysis''
      (\anainkey value anaoutkeys outvalue oldkey -> combine oldkey (keyfun anainkey value anaoutkeys outvalue))
      ana

mapAnalysis :: Keyable key => key -> AnalysisFK key a b -> AnalysisFK key [a] [b]
mapAnalysis newkey ana = mapAnalysis' (\anainkey value anaoutkeys outvalue -> newkey) ana
-}

{-
-- Simple Examples
countLoggings :: AnalysisNew [Logging] Int -- AnalysisFK (Maybe String) [Logging] Int
countLoggings = basicAnalysis ("number of loggings") length
countLoggings2 = basicAnalysis (Just $ "number of loggings") length

type AnalysisNew a b = AnalysisFK String a b

lOC :: AnalysisNew Logging Integer
lOC = basicAnalysis "Lines of code" $ toInteger .  m1_2_4

allLOC :: AnalysisNew [Logging] [Integer]
allLOC = mapAnalysis "applied over a set of logging" lOC

isLoggingFromUser:: Identity -> AnalysisNew Logging Bool
isLoggingFromUser identity = basicAnalysis ("log from " ++ show identity) (isIdentity identity)

loggingsFromUser :: Identity -> AnalysisNew [Logging] [Logging]
loggingsFromUser identity = -- applyAnalysis ("Loggings from " ++ show identity) (filter (isIdentity identity) )
   filterAna ("Loggings from " ++ show identity) (isLoggingFromUser identity)

-- Grouping analysis
grpLoggingsPerUser :: AnalysisNew [Logging] [Logging]
grpLoggingsPerUser = groupAnalysis (\log -> "loggings from " ++ (identity log)) groupPerIdentity_

noOfLoggingsPerUser :: AnalysisNew [Logging] Int
noOfLoggingsPerUser = countLoggings <.> grpLoggingsPerUser 

overallAverage :: (Real a) => AnalysisNew a Float  -- Real is really necessary
overallAverage = aggregateAnalysis' (\keys -> "Average calculated over " ++ show keys) mean

avgNoOfLoggingsPerUser :: AnalysisNew [Logging] Float
avgNoOfLoggingsPerUser = overallAverage <.> noOfLoggingsPerUser
-- short hand for: overallAverage <.> countLoggings <.> createGrpAnalysis (\id -> "loggings from " ++ id) groupPerIdentity

-- NOTE: The next (error prone) calculation / analysis combination is posible:
wrongAvgNoOfLoggingsPerUser :: AnalysisNew [Logging] Float
wrongAvgNoOfLoggingsPerUser = overallAverage <.> countLoggings
-- so it is possible to aggregate over data, that isn't grouped (multiplicated) in the first place, due to the weak analysesresult datatype.
-- solutions:
-- - check on key, resulting in a runtime warning/error,
-- - more informative key, resulting in a proper description of the non-sense result
-- - more specific result datatype, with compiler time error, folding algebra can be used :)

usersInLoggingSelection :: AnalysisNew [Logging] [Identity]
usersInLoggingSelection =
   (basicAnalysis "Unique (nubbed)" nub) . (mapAnalysis "Mapped over set" $ basicAnalysis "Identity of Logging" identity)

-}