module Statistics.DescriptiveAnalysis.PrimAC where

{-

-- Elementary AnalysisCombinators, most generic to support different specialisations
-- Optional, also import / export Statistics.DescriptiveAnalysis.PrimGroupingAC

-- example only
-- import Thesis.Helium.Logging
-- import Thesis.Grouping.Logging (groupPerSession_)

Mind:
   - dirty trick, precondition of getAnalysisFun
   - Note: possible error prone, analysis combination is possible
   - Also rel. to note: getAnaResult :: (Keyable key) => a -> AnalysisFixedKey key a b -> [b]
      it returns a list of values

-}

-- An abstract analysis
type AnalysisAbstract dt1 dt2 = dt1 -> dt2


type AnalysisInOut key value = [(key, value)]

-- An abstract key oriented analysis, key usefull for descripting the tupled value, list are used to cope with grouping and aggregating
type Analysis keya a keyb b = AnalysisAbstract [(keya, a)] [(keyb, b)]

-- other option, a seperated key analysis funciton, with functions to access the both easy
data Analysis2 keya a keyb b = Analysis2 {
                                             analysis :: a -> [b],
                                             keytransf :: keya -> [keyb]
                                         }

-- Helper functions, executing analysis

-- runAnalysisCombinators / startAnalysis
runAnalysis ::  a -> keya -> Analysis keya a keyb b -> [(keyb, b)]
runAnalysis input startkey ana = ana [(startkey, input)]

getAnaResult' :: a -> keya -> Analysis keya a keyb b -> [b]
getAnaResult' input startkey = map snd . runAnalysis input startkey


-- Helper functions, to 'get' the actual analysis function and the key transformation, describing the analysis

-- peek the result, works only properly if analysis is a singleton list, and does not depend on the key
getAnalysisFun :: Analysis keya a keyb b -> (a -> [b])
getAnalysisFun ana = (\input -> map snd $ {- head $ -} ana [(analysisError "computation", input)])

-- Get single result, only safe if analysis result is singleton list
getAnalysisFunSingle :: Analysis keya a keyb b -> (a -> b)
getAnalysisFunSingle ana =
   (\input ->
      let result = getAnalysisFun ana input
      in
         if (length result == 1) then head result
                                 else error "Error evaluating the computation function of this analysis. Computations analysis result is not a single value."
   )


-- requires no dependency from the keys transformation on the analysis (result)
getKeyTransf :: Analysis keya a keyb b -> (keya -> [keyb])
getKeyTransf ana = (\keyin -> map fst $ {- head $ -}ana [(keyin, analysisError "key transforming")])

getKeyTransfSingle :: Analysis keya a keyb b -> (keya -> keyb)
getKeyTransfSingle ana = 
   (\input ->
      let result = getKeyTransf ana input
      in
         if (length result == 1) then head result
                                 else error "Error evaluating the key transforming function of this analysis. Computations analysis result is not a single value."
   )

setAnalysis :: Analysis keya a keya a
setAnalysis = basicAnalysis id id

-- helper functions
idAnalysis :: Analysis keya a keya a
idAnalysis = id

-- combinator to just start an analysis, without doing anything, used by analyse from ResearchGeneration.hs
kickStartAnalysis :: a -> key -> [(key, a)]
kickStartAnalysis input key = runAnalysis input key idAnalysis

analysisError :: String -> a
analysisError subject =
   error $ "Error evaluating the " ++ subject ++ " function of this analysis. There is a dependence between the computation and the key transformation function."

----
-- Elementary Generic Analysis Combinators, for building/starting an analysis sequence
----

-- Most general analysis, the input value is available to the key transform function
basicAnalysis''' :: (a -> b -> keya -> keyb) -> (a -> b) -> Analysis keya a keyb b
basicAnalysis''' keyfun anafun = map (\(key, value) -> (keyfun value (anafun value) key, anafun value))

basicAnalysis'' :: (a -> keya -> keyb) -> (a -> b) -> Analysis keya a keyb b
basicAnalysis'' keyfun anafun = basicAnalysis''' (\indata _ keya -> keyfun indata keya) anafun -- map (\(key, value) -> (keyfun value key, anafun value))

basicAnalysis' :: (keya -> keyb) -> (a -> b) -> Analysis keya a keyb b
basicAnalysis' keyfun anafun = basicAnalysis'' (\_ -> keyfun) anafun

basicAnalysis_ :: keyb -> (a -> b) -> Analysis keya a keyb b
basicAnalysis_ key anafun = basicAnalysis' (\_ -> key) anafun

-- typical basicAnalysis selector, as described in thesis, adding any extra combinator does not require to rename in examples or already made implementations
basicAnalysis :: (keya -> keyb) -> (a -> b) -> Analysis keya a keyb b
basicAnalysis = basicAnalysis'

basicAnalysis_script :: (keya -> keyb) -> (a -> b) -> Analysis keya a keyb b
basicAnalysis_script keyfun valfun = map (\(key, value) -> (keyfun key, valfun value))

groupAnalysis'' :: ([[a]] -> keya -> [keyb]) -> ([a] -> [[a]]) -> Analysis keya [a] keyb [a]
groupAnalysis'' keyfun grpfun = concatMap groupAna
   where
      groupAna (orikey, list) = zip (keyfun groupedLists orikey) groupedLists
         where
            groupedLists = grpfun list

{-
-- Example
groupPerSession :: Analysis String [Logging] String [Logging]
groupPerSession = groupAnalysis''
         (\grps oldkey -> map (\(no, grp) -> oldkey ++ "; " ++ "Session" ++ show no) $ zip [0..] grps)
         groupPerSession_
-}

-- Group analysis
groupAnalysis' :: ([a] -> keya -> keyb) -> ([a] -> [[a]]) -> Analysis keya [a] keyb [a]
groupAnalysis' keyfun grpfun = groupAnalysis'' (\grps oldkey -> map (flip keyfun oldkey) grps) grpfun

-- groupPerStudent :: Analysis String [Logging] Identity [Logging]
-- groupPerStudent = groupAnalysis' (\grp oldkey -> identity $ head grp) groupPerIdentity_

groupAnalysis'_ :: (a -> keya -> keyb) -> ([a] -> [[a]]) -> Analysis keya [a] keyb [a]
groupAnalysis'_ keyfun grpfun = groupAnalysis' (\grp oldkey -> keyfun (head grp) oldkey) grpfun

-- the most typical grouping combinator
groupAnalysis :: (a -> keya -> keyb) -> ([a] -> [[a]]) -> Analysis keya [a] keyb [a]
groupAnalysis = groupAnalysis'_

groupAnalysis_sc :: (a -> keya -> keyb) -> ([a] -> [[a]]) -> Analysis keya [a] keyb [a]
groupAnalysis_sc keyfun grpfun = concatMap groupAna
   where 
      groupAna (orikey, list) = zip (map (flip keyfun orikey . head) grps) grps
         where grps = grpfun list

-- to group without describing it
groupAnalysisSilent :: ([a] -> [[a]]) -> Analysis keya [a] keya [a]
groupAnalysisSilent grpfun = groupAnalysis (const id) grpfun

-- select group analysis combinator
selectGrpAnalysis'' :: (a -> key1 -> key2) -> (a -> Bool) -> Analysis key1 a key2 a
selectGrpAnalysis'' keyfun predfun =
   map (\(orikey, value) -> (keyfun value orikey, value)
   )
   .
   filter (\(orikey, value) -> predfun value)

selectGrpAnalysis' :: (key1 -> key2) -> (a -> Bool) -> Analysis key1 a key2 a
selectGrpAnalysis' keyfun predfun = selectGrpAnalysis'' (\_ -> keyfun) predfun

-- typical select groupings combinator
selectGrpAnalysis :: (key1 -> key2) -> (a -> Bool) -> Analysis key1 a key2 a
selectGrpAnalysis = selectGrpAnalysis'

-- selectGrpAnalysis_sc :: (key1 -> key2) -> (a -> Bool) -> Analysis key1 a key2 a
-- selectGrpAnalysis_sc keyfun predfun = 
--    map (\(orikey, value) -> (keyfun orikey, value)
--    )
--    .
--    filter (\(orikey, value) -> predfun value)

-- select_exm = selectGrpAnalysis_sc (++"min. set of 5 elms") ((> 5) .length)

-- aggregation combinator
aggregateAnalysis''' :: ([a] -> b -> [keya] -> keyb) -> ([a] -> b) -> Analysis keya a keyb b
aggregateAnalysis''' keyfun aggregatefun =
   (\input -> 
      let
         keysin = map fst input
         valuesin = map snd input
         aggr_value = aggregatefun valuesin
      in
         [(keyfun valuesin aggr_value keysin, aggr_value)]
   )

aggregateAnalysis'' :: ([a] -> [keya] -> keyb) -> ([a] -> b) -> Analysis keya a keyb b
aggregateAnalysis'' keyfun aggregatefun = aggregateAnalysis''' (\values _ -> keyfun values) aggregatefun

aggregateAnalysis' :: ([keya] -> keyb) -> ([a] -> b) -> Analysis keya a keyb b
aggregateAnalysis' keyfun = aggregateAnalysis'' (\_ -> keyfun)

aggregateAnalysis = aggregateAnalysis'

aggregateAnalysis_sc :: ([a] -> [keya] -> keyb) -> ([a] -> b) -> Analysis keya a keyb b
aggregateAnalysis_sc keyfun aggregatefun = -- aggregateAnalysis'' (\_ -> keyfun) aggrfun
   (\input ->
      let
         keysin = map fst input
         valuesin = map snd input
         aggr_value = aggregatefun valuesin
      in
         [(keyfun valuesin keysin, aggr_value)]
   )

----
-- Analysis composition
----

-- infixl 6  <.>
-- (<.>) :: Analysis keyb b keyc c -> Analysis keya a keyb b -> Analysis keya a keyc c
ana2 <.> ana1 = (ana2 . ana1)

(<$>) :: Analysis keya a keyb b -> [(keya, a)] -> [(keyb, b)]
(<$>) = ($)

-- Compose a set of analyses
multiAnalyses :: [Analysis key a key a] -> Analysis key a key a
multiAnalyses anas = foldl (<.>) id anas

-- combining analysis results, only works is the analysis result type is similar
(<+>) :: Analysis keya a keyc c -> Analysis keya a keyc c -> Analysis keya a keyc c
ana1 <+> ana2 = (\input -> (ana1 input) ++ (ana2 input)
                )

-- Analysis tupling, if one of the analyses is a grouping analysis, (or the sizes of the anas differ) the elements from the bigger result set are silently dropped.
(<&>) :: Analysis keya a keyb1 b1 -> Analysis keya a keyb2 b2 -> Analysis keya a (keyb1,keyb2) (b1, b2)
ana1 <&> ana2 = (\input ->
   let ana1result = ana1 input 
       ana2result = ana2 input
   in 
      zip (zip (map fst ana1result) (map fst ana2result)) (zip (map snd ana1result) (map snd ana2result))
   )
-- ana1 <&> ana2 = (\input -> zipWith' (error "no zipwith' implemented for <&> combinator") (ana1 input) (ana2 input)                )

-- zip, througing error in case of lists with different lengths, to auxilary?
zipWith' fun (x:xs) (y:ys) = (fun x y) : (zipWith' fun xs ys)
zipWith' _ _ _ = error "Error while applying analysis parallel. Analyses are of different domain (input) size."

-- Abstraction for straight forward (injective) list manipulation analysis, filter, map, etc
listFunAna :: ((a -> b) -> [a] -> [d]) -> ([a] -> keyc -> keyd) -> (Analysis keya a keyb b) -> Analysis keyc [a] keyd [d]
listFunAna fun keyfun anafun =
   map (\(key, value) -> (keyfun value key, fun (getAnalysisFunSingle anafun) value))  -- unsafe single fun

filterAna'' :: ([a] -> keyc -> keyd) -> (Analysis keya a keyb Bool) -> Analysis keyc [a] keyd [a]
filterAna'' keyfun selectana = listFunAna filter keyfun selectana 

{- Some simpelere non- analysis like implementations are also possible
filterOnId :: String -> Analysis String [Logging] String [Logging]
filterOnId ident = basicAnalysis' (\key -> key ++ "filter on user ") (filter ((==) ident . identity))

-- or 
filterOn :: (a -> Bool) -> Analysis String [a] String [a]
filterOn fun = basicAnalysis' (\key -> key ++ "filter on ") (filter fun)

filterOnId' id = filterOn ((==) id . identity)
-}

filterAna' :: (keya -> keyb) -> (Analysis keya a keyb Bool) -> Analysis keya [a] keyb [a]
filterAna' keyfun selectana = filterAna'' (\_ -> keyfun) selectana

-- Helper function
-- lifting a analysis to a analysis on a list of the same values.
-- if analysis is a grouping analysis, the grouping will not be effective
mapAnalysis :: Analysis keya a keyb b -> Analysis keya [a] keyb [b]
mapAnalysis ana = 
   map (\(key, value) -> (head $ (getKeyTransf ana) key, concat $ map (getAnalysisFun ana) value))
   where
      mappingValue value = map (getAnalysisFun ana) value
      mappingKey   key = map (getKeyTransf ana) key

-- splitting analysis result to the groups, for easy mapping of analysis over each group
splitAnalysis :: [(key, a)] -> [[(key, a)]]   -- use a selectGrpAna?
splitAnalysis anaresult = [[x] | x <- anaresult]

-- Examples in Thesis.Research.Common for applying an analysis on each result set of an analysis in isolation

-- isolateAnalysis combinator for lifting an analysis to a context where it is applied on the group as such
-- if ana is a grouping ana, aggregation is necessary to keep the information for the group together, see ordValueCount
isolateAnalysis :: Analysis keya a keyb b -> Analysis keya a keyb b
isolateAnalysis analysis =
       concat
  <.>  map analysis
  <.>  splitAnalysis

-- Special apply combinator to apply an analysis for each result set
(<..>) :: Analysis keyb b keyc c -> Analysis keya a keyb b -> Analysis keya a keyc c
ana1 <..> ana2 = (isolateAnalysis ana1) <.> ana2

--, to flatten tupled and grouped analysis result to a list
flattenAnalysis :: ((b,a) -> key1 -> key2) -> Analysis key1 [(b, a)] key2 a
flattenAnalysis keyfun =
         basicAnalysis id (snd . head)
    <.>  groupAnalysis keyfun split

split :: [x] -> [[x]]
split list = [[x] | x <- list]

-- Example specialization for key = String
flattenAnalysis' :: Show b => Analysis String [(b, a)] String a
flattenAnalysis' =
      basicAnalysis (++"; flattend") (snd . head)
  <.> groupAnalysis (\val oldkey -> oldkey ++ "; " ++ (show $ fst val)) (split)


{-
-- Testers

logs :: [Logging]
logs = [log1, log2, log3, log4]

testRunAna :: Analysis String [Logging] keyb b -> [(keyb, b)]
testRunAna = runAnalysis' logs ""

testRunAnaSingleLog :: Analysis String Logging keyb b -> [(keyb, b)]
testRunAnaSingleLog = runAnalysis' log1 ""
-}