{-# OPTIONS_GHC -fallow-undecidable-instances #-}
module Statistics.Presentation.MarkUpDoc (
           module Statistics.Presentation.MarkUpDoc 
         , module Text.Html
         ) where


import Distribution.Compat.FilePath
import Text.Html hiding (renderHtml, renderHtml', renderTag)
import qualified Text.Html (renderHtml, renderHtml')
import Data.List (lookup, find, intersperse)
import Data.Maybe
import Auxilary(traceVal, debug,stringToFilePath)
import Text.Regex
import Data.Char

{-
----
-- MarkUpDoc: html/tex like combinators able to output html as well as tex
----

Implementation by encapsulating Text.Html combinators (requires -fallow-undecidable-instances)
Based on the idea of Skribe: http://www-sop.inria.fr/mimosa/fp/Skribe/

Todo:
- image to render combinator
- render multi (html) files
- css attachen for html output
- less html / tex only combinator for markupTable and figure, better rendering
- filter latex attrs in html rendering
- catch possible read .. :: Int parse error
- MultiRow, implemation, 1 fst cell should be empty http://www.minds.nuim.ie/~page/serendipity/index.php?/archives/43-Multirow-and-multicolumn-spanning-with-latex-tables.html
- proper new paragraph start
- sortable html table, http://kryogenix.org/code/browser/sorttable/
-}

type MarkUpDoc = Html

class (HTML a) => MARKUPDOC a where
   toMarkUpDoc     :: a -> MarkUpDoc
   toMarkUpDoc     = toHtml
   toMarkUpDocFromList :: [a] -> MarkUpDoc
   toMarkUpDocFromList xs = toHtmlFromList xs-- MarkupDoc (concat [ x | (Html x) <- map toMarkUpDoc xs])

instance (HTML a) => MARKUPDOC a where

-- different output rendering formats
data OutputFormat =
     LaTeXFormat
   | LaTeXFormatInclude   -- strips any header / document or documentclass
   | HtmlFormat
   | HtmlFormatInclude    -- strips any header / document or documentclass
   deriving (Show, Read, Eq)

showFormatExt :: OutputFormat -> String
showFormatExt LaTeXFormat           = "tex"
showFormatExt LaTeXFormatInclude    = "tex"
showFormatExt HtmlFormat            = "html"
showFormatExt HtmlFormatInclude     = "html"

-- MarkUp combinators, analog to Html variants
noMarkUp = noHtml
emptyMarkUp = primMarkUp ""
primMarkUp = primHtml
newline = br
emptylines n = concatHtml $ take n $ repeat newline
emptycell = td << spaceHtml
nocell = td << noHtml
concatMarkUp :: (MARKUPDOC a) => [a] -> MarkUpDoc
concatMarkUp = concatHtml
-- check wheter more tex in inside
isNoTeX = isNoHtml

-- Tex operators/commands, based on Piets Latex Manual, page 13
data TexCmdType =
     SingleCharCmd 
   | PlainCmd  -- with or without innerTeX; example: \large \copyright \section{SectionName \copyright\ by me} \section{SectionName \copyright{} by me}  (Mind the trailing space, used as cmd seperator)
   | PlainCmdOnlyOptions -- plain command, with no real inner content, inner content used to concat it afterwards, but not inside, example: documentclass
   | EnvCmd   -- begin / end cmd or environment; example: \begin{itemize} \item{sdf} \item{sdf} \end{itemize}
   deriving (Show, Read, Eq)

-- ref to CmdType type
teXCmdType = "texcmdtype"

-- Latex like structure/markup tags, with correct transformation in any output format, some tags/attributes only for certain outputs

-- to create document, which will properly render in html and latex,
-- or use: mydoc (docclass << thetitle << "Nice work") +++ texdocument << "content"
document :: (MARKUPDOC a, MARKUPDOC b) => a -> b -> MarkUpDoc
document preamble content =
   document' preamble [settexparam1 "[11pt]{artikel1}"] content []

-- more generic document combinator
document' preamble preAmAttrs content contentAttrs =  -- two way, for latex use preambule and html:
    header ! preAmAttrs  -- ! [preamble $ show preambule]  -- preamble attr unimplemented, but usefull for latex only tags
            << preamble -- title, and other stuff like metatags and latex settings
    +++
       body ! contentAttrs << content

docclass = header
texdocument = body

section :: (MARKUPDOC a) => Maybe String -> a -> MarkUpDoc
section ident inner =
      (if isJust ident
         then
            anchor ! [name (fromJust ident), formatonly $ show HtmlFormat] << noHtml
         else
            noHtml
      )
      +++
         h3 ! (if isJust ident then [label $ fromJust ident] else [])
                  << inner

subsection ident inner =
      (if isJust ident
         then
            anchor ! [name (fromJust ident), formatonly $ show HtmlFormat] << noHtml
         else
            noHtml
      )
      +++
         h4 ! (if isJust ident then [label $ fromJust ident] else [])
                  << inner
                  
subsubsection ident inner =
      (if isJust ident
         then
            anchor ! [name (fromJust ident), formatonly $ show HtmlFormat] << noHtml
         else
            noHtml
      )
      +++
         h5 ! (if isJust ident then [label $ fromJust ident] else [])
                  << inner

-- NB: div specially used to render as \caption in latex,
figurecaption = thediv -- ! [thestyle "font-weight: bold"] 

markUpTable2D :: (MARKUPDOC a, MARKUPDOC b) => Maybe String -> String -> [String] -> [(a,[b])] -> MarkUpDoc
markUpTable2D ident title cols tabdata =
   let reflabel = maybe (stringToIdent title) id ident
   in
      -- html table
      anchor ! [name reflabel , formatonly $ show HtmlFormat] << noHtml
      +++
      table ! [border 1, cellspacing 0, cellpadding 0, formatonly $ show HtmlFormat]
         << (
            tr << (emptycell +++ (concatMarkUp $ map (\c -> td << c) cols))
            +++
               (concatMarkUp $ map (\(x, ys) -> tr << (td << x +++ (concatMarkUp $ map (\c -> td << c) ys))) tabdata)
            )
      +++
      figurecaption ! [label reflabel, formatonly $ show HtmlFormat] << title
      +++
      paragraph ! [formatonly $ show HtmlFormat] << noMarkUp
      +++
      -- latex table in figure env (inner elm are ignored in other output formats)
      figure_env ! [formatonly $ show LaTeXFormat] <<
         (table ! [border 1, cellspacing 0, cellpadding 0]
            << (
               tr << (emptycell +++ (concatMarkUp $ map (\c -> td << c) cols))
               +++
                  (concatMarkUp $ map (\(x, ys) -> tr << (td << x +++ (concatMarkUp $ map (\c -> td << c) ys))) tabdata)
               )
            +++
            figurecaption ! [label reflabel] << title
         )

markUpTable1D :: (MARKUPDOC a, MARKUPDOC b) => Maybe String -> String -> String -> [(a,b)] -> MarkUpDoc
markUpTable1D ident title col tabdata =
   markUpTable2D ident title [col] (map (\(x,y) -> (x,[y])) tabdata)

-- to display a generated image
figure :: Maybe String -> String -> (FilePath,FilePath) -> MarkUpDoc
figure ident title (baseurl,relurl) {- baseurl for LateX -}=
   let reflabel = maybe (stringToIdent title) id ident
   in
      -- html figure / image
      anchor ! [name reflabel , formatonly $ show HtmlFormat] << noHtml
         +++ 
      image ! [src relurl, formatonly $ show HtmlFormat]
         +++
      figurecaption ! [label reflabel, formatonly $ show HtmlFormat] << title
         +++
      paragraph ! [formatonly $ show HtmlFormat] << noMarkUp
         +++
      -- latex image
      figure_env ! [formatonly $ show LaTeXFormat] <<
                        (image ! [settexparam1 "[width=8cm]", src (joinFileName baseurl relurl)]
                           +++
                         figurecaption ! [label reflabel] << title
                        )

stringToIdent = stringToFilePath -- filter (not . (== ' '))
-----------------------------------------------------------------------------------------------------------
-- Latex Rendering Implementation
-----------------------------------------------------------------------------------------------------------

-- for creating latex only tags
textag :: TexCmdType -> String -> MarkUpDoc -> MarkUpDoc
textag tagt name  htmls = Html [
      HtmlTag {
              markupTag = name,
              markupAttrs = [texcmdtype (show tagt)],  -- when using concrete tag add: formatonly $ show LaTeXFormat as attr
              markupContent = htmls }
      ] 

-- typical latex commands
frontmatter, maketitle :: MarkUpDoc
frontmatter = textag PlainCmdOnlyOptions "frontmatter" noMarkUp
maketitle = textag PlainCmdOnlyOptions "maketitle" noMarkUp
figure_env = textag EnvCmd "figure"

-- table only tags, only used to renderLateXtable
hline      = textag PlainCmdOnlyOptions "hline" noMarkUp
endtabline = textag SingleCharCmd ("\\\\\n")  noMarkUp
tabtab     = textag SingleCharCmd       "&"     noMarkUp

-- extra attribute to handle different rendering, mainly for LateX rendering
-- Mind: keep name the same and the key value, to avoid name clash with existing html attributes
-- \<attr>{<value>}
label = strAttr "label"  -- usefull to set extra functionality for latex support for referencing, set by identifier

-- show by value
settexparam1 = strAttr "settexparam1"
settexparam2 = strAttr "settexparam2"
settexparam3 = strAttr "settexparam3"
prodtagcont  = strAttr "prodtagcont"  -- special case, to prevent dubble processing, of latex rendering (otherwise stringToLatex would run twice on the content

-- only used to handle rendering diff between output types
formatonly = strAttr "formatonly"  -- to set attribute rendering specific
texcmdtype = strAttr teXCmdType  -- to set TexCmdType

texattrs = ["settexparam1", "settexparam1", "settexparam1", "prodtagcont", "formatonly"]

-- html tag to tex tag, pre rendering procedure
-- html2TeX :: MarkUpDoc -> MarkUpDoc
html2TeX :: HtmlElement -> Maybe HtmlElement
html2TeX tag@(HtmlTag
              { markupTag = name,
                markupContent = innertex,
                markupAttrs = markupAttrs })
      =     case name of
               "HEAD"
                  -> Just $ HtmlTag
                        { markupTag = "documentclass",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmdOnlyOptions):markupAttrs}
               "TITLE"
                  -> Just $ HtmlTag
                        { markupTag = "title",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmd):markupAttrs}
               "BODY"
                  -> Just $ HtmlTag
                        { markupTag = "document",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show EnvCmd   ):markupAttrs}
               "NO-ENV"  -- special tag for encapsulating elms in not shown elm, for showing include output
                  -> Just $ HtmlTag
                        { markupTag = "",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show SingleCharCmd   ):markupAttrs}
               "H1" -> Nothing
               "H2" -> Just $ HtmlTag
                        { markupTag = "chapter",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmd):markupAttrs}
               "H3" -> Just $ HtmlTag
                        { markupTag = "section",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmd):markupAttrs}
               "H4" -> Just $ HtmlTag
                        { markupTag = "subsection",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmd):markupAttrs}
               "TABLE" -> Just $ (renderLatexTable tag)
--               "TR" / "TD"/ "TH" handled by renderLateXTable
               "BR" -> Just $ HtmlTag
                        { markupTag = "newline",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmdOnlyOptions):markupAttrs}
               "P" -> Just $ HtmlTag
                        { markupTag = "newline",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmdOnlyOptions):markupAttrs}
               "DIV" -> Just $ HtmlTag
                        { markupTag = "caption",
                          markupContent = innertex,
                          markupAttrs = (texcmdtype $ show PlainCmd):markupAttrs}
--               "CAPTION"  -- only to be used in table tag
               "IMG" -> let imgref = maybe "no image reference specified" id (lookupAttrVal "SRC" markupAttrs)
                        in Just $ HtmlTag {
                                 markupTag = "includegraphics",
                                 markupContent = noHtml,
                                 markupAttrs = (texcmdtype $ show PlainCmdOnlyOptions):markupAttrs++[(settexparam3 ("{"++imgref++"}"))]                                  }
               otherwise ->
                     if (isJust $ lookupAttrVal teXCmdType markupAttrs)
                     then Just tag
                     else if (isJust $ lookupAttrVal "formatonly" markupAttrs)
                          then Nothing
                          else if debug then error $ "Error: Non supported tag (" ++ (show $ markupTag tag) ++ ") being rendered as LateX. This can also be the result of incorrectly combinating html tags."
                                        else Nothing
html2TeX (HtmlString str) = Just (HtmlString (id $ undoHtmlSpecial str))


-- Latex table (tabular) rendering
renderLatexTable :: HtmlElement -> HtmlElement
renderLatexTable tag@(HtmlTag
              { markupTag = name,
                markupContent = content@(Html elms),
                markupAttrs = attrs }) =
   let
      borderset = maybe False (>0) $ lookupAttrIntVal "BORDER" attrs

      trs :: [HtmlElement]
      trs = filter (isTag "TR") elms
      tdths_filter = filter (\t -> isTag "TD" t || isTag "TH" t) 
      no_tr = length trs
      no_td =
         if (no_tr > 0)
         then sum $ map (getColSpan) $ tdths_filter $ getHtmlElements $ markupContent $ head trs
         else if debug then innertableerror
                       else 0
      
      getColSpan :: HtmlElement -> Int
      getColSpan = foldHtmlElement (const 0, (\_ tdattrs _ -> maybe 1 id $ lookupAttrIntVal "COLSPAN" tdattrs))

      innertableerror = error "No row present in table inner content"
      -- table setting, alignment of columns, default {|l|r|r|..|} (rest right aligned)
      settexparam2text = "{|" ++ (concat $ (intersperse "|") (replicate (min no_td 1) "l" ++ replicate (no_td-1) "r")) ++ "|}"
      getHtmlElmFromTag :: HtmlElement -> [HtmlElement]
      getHtmlElmFromTag = getHtmlElements . markupContent
      
      renderLateXTableTDS :: Int -> HtmlElement -> String
      renderLateXTableTDS n trtag =
         let
            tds = tdths_filter $ getHtmlElmFromTag trtag
            colspan attrs = maybe 0 id $ lookupAttrIntVal "COLSPAN" attrs
         in
            concat $ intersperse tab $
               map (\tdtag ->
                     let colspanno = (colspan $ markupAttrs tdtag)
                     in
                        if colspanno > 1
                        then ("\\multicolumn{" ++ show colspanno ++ "}" ++ "{|c|}{" ++ (renderTeXString n $ getHtmlElmFromTag tdtag) ++ "}")
                        else (renderTeXString n $ getHtmlElmFromTag tdtag)
                   ) $ tds
      hl = "\\hline\n"
      tab = "& "
      el = "\\\\"
      tableContent = 
            if (no_tr == 0)   -- should use borderset
               then (hl ++ el ++ hl)
               else if (no_tr == 1)
               then (hl ++ (renderLateXTableTDS 0 $ head trs) ++ el)
               else -- (no_tr > 1) = -- show table, only multicolumn supported
                  (hl ++ (renderLateXTableTDS 0 $ head trs) ++ el ++ "\n" ++
                      hl ++ (concat $ intersperse (el ++ "\n") $ map (renderLateXTableTDS 0) $ tail trs) ++ el ++ "\n" ++ hl)
   in HtmlTag
         { markupTag = "tabular",
           markupContent = emptyMarkUp,
           markupAttrs = (settexparam1 "[t]"):(settexparam2 settexparam2text):(prodtagcont tableContent):(texcmdtype $ show EnvCmd):attrs
         }

-- eval TeX to string, to embed in document
renderTeXString :: Int -> [HtmlElement] -> String
renderTeXString n texdoc = foldr (.) id (map (renderLateX'Filter n) texdoc) "\n"

renderMarkUpDoc :: OutputFormat -> MarkUpDoc -> String
renderMarkUpDoc format doc =
   case format of
      LaTeXFormat          -> renderLateX {-$ filterDoc LaTeXFormat -}doc
      LaTeXFormatInclude   -> renderLateX $ {-filterDoc LaTeXFormat $ -}stripHeader doc  -- ignore HEAD
      HtmlFormat           -> renderHtml doc
      HtmlFormatInclude    -> renderHtml $ stripHeader doc  -- ignore preamble

renderHtml doc = Text.Html.renderHtml $ filterDoc HtmlFormat doc

-- only render element with attribute formatonly set to HtmlFormat or no setting of this attribute at all
-- and remove any tex related attributes
filterDoc :: OutputFormat -> Html -> Html
filterDoc fmt = Html . (map hasCorrHtmlChilds). (filter hasCorrAttrFormat) . getHtmlElements
   where
      hasCorrAttrFormat :: HtmlElement -> Bool
      hasCorrAttrFormat =
         foldHtmlElement (const True, (\_ attrs _ -> maybe (True) (==show fmt) (lookupAttrVal "formatonly" attrs)))
      hasCorrHtmlChilds = foldHtmlElement (HtmlString,
                              (\name attrs content -> HtmlTag name ({-filter (\atr -> (not . and . map (isAttr texattrs)) -}attrs) (filterDoc fmt content)))

stripHeader :: Html -> Html
stripHeader doc = renameBodyToNo_Env $ Html . filter (not . isTag "HEAD") $ getHtmlElements doc
   where
      renameBodyToNo_Env :: Html -> Html
      renameBodyToNo_Env html = Html $ map renameTagBODY' $ getHtmlElements html

      renameTagBODY' (HtmlString str) = HtmlString str
      renameTagBODY' tag | markupTag tag == "BODY" = HtmlTag {markupTag = "NO-ENV", markupAttrs = markupAttrs tag, markupContent = markupContent tag}
                         | otherwise                = tag

renderLateX :: (MARKUPDOC doc) => doc -> String
renderLateX doc =
      renderMessageLateX ++
         foldr (.) id ( map (renderLateX'Filter 0)
                           -- $  (mapMaybe html2TeX  -- added fix to make htmltags sort of presentable in tex, functions as filter as well, filter used on 2 places, also to filter on inner contents
                                 -- (getHtmlElements (tag "HTML" << doc)))) "\n"
                                 (getHtmlElements ({-toHtml-} latexdoc))) "\n"
   where latexdoc = filterDoc LaTeXFormat $ toHtml doc

renderMessageLateX = "" -- "% Rendered using the Haskell MarkUpDoc Library v0.1\n"

-- point to do tag transformation
renderLateX'Filter :: Int -> HtmlElement -> ShowS
renderLateX'Filter n tag = maybe id (renderLateX' n) $ html2TeX tag

renderLateX' :: Int -> HtmlElement -> ShowS
renderLateX' _ (HtmlString str) = (++) (stringToLateXString str)
renderLateX' n (HtmlTag
              { markupTag = name,
                markupContent = innertex,
                markupAttrs = markupAttrs })
      =
        if isNoTeX innertex -- && elem name validLateXTags
        then renderLateXTag True defltCmd name markupAttrs n
        else (renderLateXTag True defltCmd name markupAttrs n
             . foldr (.) id (map (renderLateX'Filter (n+2))
                     (getHtmlElements innertex))
             . renderLateXTag False defltCmd name markupAttrs{-[]-} n)
        where defltCmd = PlainCmd  -- ex: \large{}

undoHtmlSpecial :: String -> String
undoHtmlSpecial htmlstr =
   (foldr (.) id
         (map (\(strout, strin) -> (\input -> subRegex (mkRegex strout) input strin)) replacements)
   ) htmlstr
   where
      replacements = [
           ("&amp;" , "&")
         , ("&lt;"  , "<")
         , ("&gt;"  , ">")
         , ("&quot;", "\"")
         , ("&nbsp;", " " )
         ]

-- Piet Latex Handleiding, page 19
stringToLateXString :: String -> String
stringToLateXString str = concatMap fixChar str
   where
      fixChar '$' = "\\$"
      fixChar '&' = "\\&"
      fixChar '%' = "\\%"
      fixChar '#' = "\\#"
      fixChar '_' = "\\_"
      fixChar '{' = "\\{" --(chr $ ord '\\'):"}" -- :"{"
      fixChar '}' = "\\}"
-- untested
--      fixChar '"' = "''"
--      fixChar '~' = "\tilde "  
--      fixChar '^' = "''"
--      fixChar '\' = "''"
--      fixChar '|' = "''"
      fixChar '<' = "$<$"
      fixChar '>' = "$>$"
      fixChar c   = [c]

-- This prints the Tex 'Tags' based on their provided tag type.
renderLateXTag :: Bool -> TexCmdType -> String -> [HtmlAttr] -> Int -> ShowS
renderLateXTag open dfltcmdform name markupAttrs n r =
      (showCmd (getCmdType dfltcmdform markupAttrs) open markupAttrs) ++ {-rest markupAttrs ++-}  r
  where
      getCmdType dflt attrs =
         let cmdType = lookupAttrVal teXCmdType attrs
         in if isJust cmdType
            then fromJust cmdType
            else if debug then
                     error "Error while rendering LateX, the case should not have occured. There is no CmdType rendering available for this type" -- show dflt
                 else
                     "unsupported tag, no cmd type available to render this in LateX"

      showCmd cmdtype open_close attrs
         | cmdtype == show SingleCharCmd = if open then name else "" -- for td trans
         | cmdtype == show PlainCmdOnlyOptions
                                  = if open
                                    then ('\\':name) ++ (rest attrs) ++ "" ++ nl -- innercontent render in parent function, final cmd options sign "\ " or {} by other cmd type
                                    else ""
         | cmdtype == show PlainCmd = if open
                                    then ('\\':name) ++ "{"  -- safegaurd to not lose meaning of next space
                                    else ("}" ++ rest attrs) -- latter {} safegaurd to not lose meaning of next space, removed to make use of plainCmdOnlyOptionsStr for this kind of cmds
         | cmdtype == show EnvCmd    = if open
                                    then (((nl ++ "\\begin{") ++ name) ++ "}") ++ (rest attrs) ++ nl
                                    else (((nl ++ "\\end{") ++ name) ++ "}") ++ nl
         | otherwise              = if debug
                                    then
                                       error "Error while rendering LateX, the case should not have occured"
                                    else
                                       "% No rendering support for tag " ++ name ++ ". You probably used tags unsupported in LateX (or unimplemented by the rendering)"

      nl = "\n" ++ replicate (n `div` 8) '\t' 
                ++ replicate (n `mod` 8) ' '

      rest attr = concat (map showPair attr)

      -- bit hacky
      showPair :: HtmlAttr -> String
      showPair (HtmlAttr tag val)
         | tag == "formatonly" = ""
         | tag == teXCmdType = ""
         | tag == "label"  = "\\" ++ tag ++ "{" ++ val ++ "}"
         | tag == "settexparam1" = val
         | tag == "settexparam2" = val
         | tag == "settexparam3" = val
         | tag == "prodtagcont"  = val
         | otherwise = ""

----
-- Helper functions
----

-- HtmlElement fold / map 
foldHtmlElement :: (String -> a, String -> [HtmlAttr] -> Html -> a) -> HtmlElement -> a
foldHtmlElement (htmlstr, htmltag) (HtmlString str) = htmlstr str
foldHtmlElement (htmlstr, htmltag) (HtmlTag name attr content) = htmltag name attr content

isTag :: String -> HtmlElement -> Bool
isTag tagname (HtmlString _) = False
isTag tagname tag | markupTag tag == tagname = True
                  | otherwise                = False

isAttr :: String -> HtmlAttr -> Bool
isAttr attrname (HtmlAttr name value) = attrname == name

-- lookup on attribute value
lookupAttrVal :: String -> [HtmlAttr] -> Maybe String
lookupAttrVal key attrs = lookup key $ makeAssListHA attrs

makeAssListHA = map (\(HtmlAttr tag val) -> (tag,val))

lookupAttrIntVal :: String -> [HtmlAttr] -> Maybe Int
lookupAttrIntVal key attrs = maybe Nothing (\str -> Just (read str :: Int)) $ lookupAttrVal key attrs

----
-- Examples use
----
{-
writeExFile = writeFile "example.html" . (renderMarkUpDoc HtmlFormat)
writeTeXFile = writeFile "example.tex" . (renderMarkUpDoc LaTeXFormat)
-- main = writeFile "example.html" $ renderHtml completedoc

testdoc :: MARKUPDOC a => a -> MarkUpDoc
testdoc =
     document title
     where
         title = thetitle  << "My first MarkupDoc generated Home Page"

-- testsec1 :: Html
testsec1 = section (Just "Foo") << "Foo goes home"
textsec1 = "for a cup of tea" -- +++ reference "foo" << "footje"

testsec2 = section (Just "Bar") << "Bar goes away"
textsec2 = "for a cup of coffee" -- +++ reference "bar" << "barretje"

testsec3 = section Nothing << "What ever" -- (Just "nietmeer")
textsec3 = "Lala lala." -- +++ reference "bar" << "barretje"

completedoc =
   testdoc <<
      (   testsec1 +++ textsec1 +++ (p << noHtml) +++ {-table ! [border 1]<< tr << (td << "Ted" +++ td << "Brak")-}  dirTable [("markup1.html", 13000),("markup2.html",40000)]
      +++ testsec2 +++ textsec2 +++ emptylines 20
      +++ testsec3 +++ textsec3 )

dirTable :: [(String, Int)] -> MarkUpDoc
dirTable inputs =
   let
      -- getFromLists i = (map . map) (flip (!!) i)
      eta sz speed = flip (++) "s" $ show (floor (fromIntegral sz / speed))
      filerow :: String -> Int -> Html
      filerow fn fs = tr << (td << fn +++ td << (show fs ++ "k")
         +++ td << eta fs 1024
         +++ td << eta fs  512
         +++ td << eta fs 10240
         )
   in
      table ! [border 1, cellspacing 0]
         <<
            (
               tr << ((nocell ! [colspan 2]) +++ (th ! [colspan 3] << "Estimated download times"))
               +++
               tr << (nocell  +++ (th << "Size") +++ (th << "56kb") +++ (th << "512kb") +++ (th << "10mb"))  -- beside probl. also works
               +++
               (concatHtml $ map (\(file, size) -> filerow file size) inputs)
            )
-}