module Statistics.SummaryMeasures where

import Data.List ((!!), genericLength, genericIndex, sort, group)
import Data.Maybe
import Data.Typeable
import Auxilary(showMaybe, fmapTupleSnd)
import Grouping
import Statistics.BasicTypes

----
-- Statistical functions, focussed on calculating summary values
----

-- frequency :: Enum b => [a] -> b
-- frequency = toEnum . length 

frequency :: (Num b) => [a] -> b
frequency = genericLength -- instead of length, delivers num class, but is slower

numberOf = genericLength 
sequenceSize = genericLength 
size = genericLength 

mean :: Real a => [a] -> OptData Float
mean = average

-- old
--mean :: (Enum a, Fractional a) => [a] -> a 
--mean list = (sum list) / (toEnum $ frequency list)   -- (toEnum $ frequency list) toEnum not necessary anymore

average       ::  Real a => [a] -> OptData Float
average []    =   NaN
average list  =   Value $ realToFrac (sum list) / realToFrac (length list)

-- minimum :: (Ord a) => [a] -> a
-- minimum -- from Prelude

-- maximum :: (Ord a) => [a] -> a
-- maximum -- from Prelude

medianOrd :: (Ord a) => [a] -> OptData (Either a (a,a))
medianOrd [] = NaN
medianOrd xs =
   case (odd ln) of
      False -> Value ( if ( (sorted !! xth) == (sorted !! (xth - 1) ))
                        then Left (sorted !! xth)
                        else Right ((sorted !! (xth - 1), (sorted !! xth)))
                     )
      True  -> Value $ Left $ sorted !! xth
   where
      ln = length xs
      xth = div ln 2
      sorted = sort xs

--median implementation, always returns a fractional, work on all real members
--median ::(Fractional b, Real a) => [a] -> Maybe b
median ::(Real a) => [a] -> OptData Float
median [] = NaN
median xs =
   case (odd ln) of
      False -> Value $ realToFrac (sorted !! xth + sorted !! (xth - 1)) / 2  -- the latter divide, makes output fractional, ordinal input would require different impl
      True  -> Value $ realToFrac $ sorted !! xth
   where
      ln = length xs
      xth = div ln 2
      sorted = sort xs


-- Code from: /users/pkeeken/local/apps/helium-1.2/src/heliumNT/src/staticanalysis/staticchecks/StaticCheck.ag / hs (bastiaan)
          
mode :: Ord a => [a] -> OptData a -- Value ... IF any of the elements is more common
mode xs = 
   case filter ((== maxFreq) . snd) fs of
      [(x, _)] -> Value x
      _ -> NaN
   where
      maxFreq = maximum (map snd fs)
      fs = frequencies xs

-- Code from: /users/pkeeken/local/apps/helium-1.2/src/heliumNT/src/staticanalysis/staticchecks/StaticCheck.ag / hs (bastiaan)
-- occurences counter
frequencies :: Ord a => [a] -> [(a, Int)]
frequencies = map (\ys -> (head ys, length ys)) . group . sort


modeMultiple :: Ord a => [a] -> OptData [a]
modeMultiple xs 
   | length maxs > 0 && length maxs < 5 = Value $ map fst maxs
   | otherwise                          = NaN
   where
      maxs = filter ((== maxFreq) . snd) fs 
      maxFreq = maximum (map snd fs)
      fs = frequencies xs

modeNomMultiple :: Eq a => [a] -> OptData [a]
modeNomMultiple xs 
   | length maxs > 0 && length maxs < 5 = Value $ map fst maxs
   | otherwise                          = NaN
   where
      maxs = filter ((== maxFreq) . snd) fs 
      maxFreq = maximum (map snd fs)
      fs = map (\ys -> (head ys, length ys)) $ groupAllByEq (==) xs

-- Faster combined implementation (freq, mean, min, max, median, mode)
-- giveBasicStats :: (Real a, Fractional b1, Fractional b) => [a] -> (Int, OptData (b, a, a, b1, OptData a))
giveBasicStats :: (Real a, Fractional b) => [a] -> (Int, OptData (b, a, a, Float, OptData a))
giveBasicStats list =
   if freq > 0
   then (freq, Value (avg, head sorted, last sorted, fromValue $ median list, modus)) --, mode list
   else (freq, NaN)
   where 
      sorted = sort list
      freq = length list
      avg = realToFrac (sum list) / realToFrac freq
   {- med = case (odd freq) of
               False -> realToFrac (sorted !! xth + sorted !! (xth - 1)) / 2
               True  -> realToFrac $ sorted !! xth
         where xth = (floor $ (realToFrac freq)/2) - 1 -- list first element is at 0th position in the list when !! is used -}
      modus = case filter ((== maxFreq) . snd) fs of  
                 [(x, _)] -> Value x
                 _ -> NaN
         where 
            maxFreq = maximum (map snd fs)
            fs = map (\ys -> (head ys, length ys)) . group $ sorted

-- relativeCount :: [(a, Int)] -> [(a,Float)]
relativeCount :: (Real b, Real c, Fractional c) => [(a, b)] -> [(a,c)]
relativeCount ordcount =
   let total = sum $ map snd ordcount
   in
      map (fmapTupleSnd (flip percentage total))
          ordcount

calcFractions :: (Real a, Real b, Fractional b) => [a] -> [b]
calcFractions list =
   -- map (flip calcFraction (sum list)) list
   map (flip percentage (sum list)) list

calcFraction :: (Real a, Real b, Fractional b) => a -> a -> b
calcFraction _ 0 = error "Error: Division by zero, not able to calculate fraction."
calcFraction x total = (realToFrac x) / (realToFrac total)
-- calcFraction x total = realToFrac (round $ 1000* (realToFrac x) / (realToFrac total)) / 10

percentage  ::  (Real a, Real b, Fractional b) => a -> a -> b
percentage x total = realToFrac (round (calcFraction (1000 * x) total)) / 10