module Thesis.Args where

import System.Console.GetOpt
import System.Directory
import System.Exit
import System.IO.Unsafe
import System.Time
import Control.Monad
import Data.Maybe
import Data.List
-- import Data.Typeable

import Auxilary
import Thesis.AnalysisPredefined (predefAnalyses)
import qualified Statistics.Presentation.MarkUpDoc as MUD (OutputFormat (..)) 

version = "1.0"

versionBuildDate :: IO String
versionBuildDate = do
   time <- getClockTime
   return $ version ++ " Build date: " ++ (calendarTimeToString $ toUTCTime time)
   
appname = "analyze"

outputformat_default = MUD.HtmlFormat

-- debug = False
data RunningMode = GHC | GHCI | Other

processArgs :: [String] -> IO ([Option], String)
processArgs args =
   let (options, arguments, errors) = getOpt Permute (optionDescription False) args
           -- "False" to be able to parse the ambiguous options for choosing an analysis research
       optShowPrefDefAnalysesDescr  = ShowAllPredefAnalysesDescr `elem` options
   in
      if ( not (null errors) || length arguments /= 1 )
         then
            do
               versionBD <- versionBuildDate
               putStrLn $ "Helium Logging Analysis " ++ versionBD
               putStrLn (usageInfo ("Usage: " ++ appname ++ " [options] logfile")
                        (optionDescription optShowPrefDefAnalysesDescr))
               when debug $ putStr ("Cmdline opts debug: " ++ show options ++ "\n" ++ show arguments ++ "\n" ++ show errors ++ "\n")
               exitWith (ExitFailure 1)
      else
         if (isOutputFilePathOption `elemBy` options)
            then
               do 
                  let outputPath = getValIfOption isOutputFilePathOption options
                  pathExists <- doesDirectoryExist outputPath
                  when (not pathExists ) (
                        do
                           putStrLn $ "Output path does not exist"
                           exitWith (ExitFailure 1)
                           ) 
                  return (options, (head arguments))
            else
               return (options, (head arguments))
   where
      optionDescription optShowPrefDefAnalysesDescr =
            {-
               -- Option descriptions:
               data OptDescr a = Option [Char] [String] (ArgDescr a) String
            -}
            [  -- Option "i" ["input-file"] (ReqArg InputFile "") "input file logfile"
              Option "a" ["analysis-reference"] (ReqArg (PredefinedAnalysis . read) "0") "run predefined analysis over input"
            , Option "A" ["all-analyses"] (NoArg RunAllAnalyses) "run all predefined analyses"
            , Option "X" ["show-predefined-analyses"] (NoArg ShowAllPredefAnalysesDescr) "show description for all predefined analyses"
--            , Option "f" ["analysis-file"] (ReqArg AnalysisFile "") "analysis script or executable"
            , Option "o" ["output-path"] (ReqArg OutputFilePath "/tmp/") "output path for analysis results"
            , Option "m" ["output-format"] (ReqArg (OutputFormat) (show MUD.HtmlFormat)) "output format, LaTeXFormat or HtmlFormat (default)"
--            , Option "v" ["verbose"] (NoArg Verbose) "extra context information of analysis"
            , Option "h" ["help"] (NoArg Help) "shows this information"
            ] ++
            if not optShowPrefDefAnalysesDescr then [] else 
            (showPredefAnalysesDescrs predefAnalyses) -- :: [OptDescr Option]

-- showPredefAnalysesDescrs :: [(Char, a)] -> [OptDescr Option]
showPredefAnalysesDescrs anas = map mkOptDescr $ zip [0..] anas
   where
      mkOptDescr (no, (descr, _)) =
         Option ("a") [{-removeSpaces descr-}] (ReqArg (PredefinedAnalysis . read) (show no)) ("Research " ++ show no ++ ": " ++ descr)

data Option
    = InputFile String
    | OutputFilePath String
    | OutputFormat String -- MUD.OutputFormat
    | AnalysisFile String
    | PredefinedAnalysis Int
    | ShowAllPredefAnalysesDescr
    | RunAllAnalyses
    | Verbose
    | Help 
 deriving (Eq, Show) -- added pvk


isPredefinedAnalysisOption :: Option -> Bool
isPredefinedAnalysisOption x =
   case x of
      PredefinedAnalysis _ -> True
      otherwise -> False

isOutputFilePathOption :: Option -> Bool
isOutputFilePathOption x =
   case x of
      OutputFilePath _ -> True
      otherwise -> False

isOutputFormatOption :: Option -> Bool
isOutputFormatOption x =
   case x of
      OutputFormat _ -> True
      otherwise -> False

-- get values from analysis <number> options
getAnaVal opt =
   case opt of
      PredefinedAnalysis val -> Just val
      otherwise -> Nothing

-- get values from analysis <reference> options, -A / --allanalyses dominates over selecting a certain analysis
getPredefinedAnalysis:: [Option] -> [Int]
getPredefinedAnalysis options =
   if (RunAllAnalyses `elem` options)
      then
         [0.. (length predefAnalyses - 1)]
   else
      mapMaybe getAnaVal $ filter isPredefinedAnalysisOption options

-- get values from option <string> options
--getVal :: forall a . Typeable a => [Option] -> a
getVal :: Option -> Maybe String
getVal opt =
   case opt of
--      PredefinedAnalysis val -> Just val
      OutputFilePath val -> Just val
      AnalysisFile val -> Just val
      otherwise -> Nothing

getValIfOptions :: (Option -> Bool) -> [Option] -> [String]
getValIfOptions pred options = mapMaybe getVal $ filter pred options

-- get value of first option in parameter list
getValIfOption :: (Option -> Bool) -> [Option] -> String
getValIfOption pred options =
   let optionValues = mapMaybe getVal $ filter pred options
   in
      case (length optionValues > 0) of
         True -> head optionValues   -- Just
         False -> error "Option value not in parameter list"   -- Nothing

-- special function for handling outputformat, throws warning if parameter not recognized
parseFmtOpt :: String -> IO MUD.OutputFormat
parseFmtOpt optval | optval == show MUD.HtmlFormat  = return MUD.HtmlFormat
                   | optval == show MUD.HtmlFormatInclude  = return MUD.HtmlFormatInclude
                   | optval == show MUD.LaTeXFormat = return MUD.LaTeXFormat
                   | optval == show MUD.LaTeXFormatInclude = return MUD.LaTeXFormatInclude
                   | otherwise                  =
                           do {-
                              putStrLn $
                                 "No parse for output format, either specify -m " ++ show MUD.HtmlFormat ++ " or -m " ++ show MUD.LaTeXFormat
                              return outputformat_default
                              -}
                              error $
                                 "No parse for output format, either specify -m " ++ show MUD.HtmlFormat ++ " or -m " ++ show MUD.LaTeXFormat ++ ". Concat \"Include\" to include the output into a document."
