module Thesis.Helium.LogfileParser where 

import Thesis.Helium.Logging

import Text.ParserCombinators.Parsec
import System.Time
import Auxilary

type LogfileLine = [Logging] -- simplified logging, empty if commented line

tstLoggingLine :: String
tstLoggingLine = "group80:L:1.2 (Tue Jan 27 14:56:04 WEST 2004)::group80/2004-01-03@12_46_12_871/Hello.hs"

parseLogfile' :: String -> IO (Either ParseError [Logging])
parseLogfile' logfile = do result <- parseFromFile pLoggings logfile
                           return result

-- For parsing the complete logfile
parseLogfile :: FilePath -> IO [Logging]  -- formerly getLogs
parseLogfile logfile =
    do result <- parseLogfile' logfile
       case (result) of
          Left err  -> {-do print err
                          return []-}
                       do putStr "Parse error: "
                          error $ show err
          Right xs  -> return $ map (addBaseDir . getDirName $ logfile) xs
   where
      getDirName = -- reverse . tail . dropWhile (/= '/') . reverse  -- should be splitFilePath (Cabal 6.4)
         fst3 . splitFilePath 
      addBaseDir basedir log = Log (username log)
                                   (phase log)
                                   (heliumVersion log)
                                   (Path basedir (userDir path) (loggingDir path) (sourceFile path))
                                   (logDate log)
                                   (staticErrorCodes log)
                                   (extraLoggingInfo log)
         where
            path = logPath log

{- obsolete, just for module testing
parseLogfileTest :: String -> IO ()
parseLogfileTest logfile = do result <- parseFromFile pLoggings logfile 
                              case (result) of
                                 Left err  -> print err
                                 Right xs  -> print (xs)
                              return ()

parseLogfileFunTest :: (Show a) => String -> ([Logging] -> a) -> IO ()
parseLogfileFunTest logfile fun = do result <- parseFromFile pLoggings logfile 
                                     case (result) of
                                        Left err  -> print err
                                        Right xs  -> print (fun xs)
                                     return ()
-}

parseLogfileLine :: String -> Either ParseError LogfileLine
parseLogfileLine line = parse pLine "line from logfile" line

pLoggings :: Parser DataSet
pLoggings = do lines <- sepEndBy pLine newline
               return (concat lines) -- map (\(LoggingLine logging) -> logging) (filter (isLogging) lines))

pLine :: Parser [Logging]
pLine = do logging <- pLogging
           return [logging]
        <|>
        do comment <- pComment 
           return []

pComment :: Parser String
pComment = do char '#'
              comment <- many (noneOf "\n")
              return comment
           <?> "comment line"

{- 
   Parser for single logging and it elements, except extensions
-}
pLogging :: Parser Logging
pLogging = do user <- pIdentifier
              separator
              (phase, errors) <- pPhase
              separator
              version <- pVersion
              separator
              date <- pBuildDate
              separator
              path <- pPath
              return (Log (Username user) phase (toHelVer version date) path (stringToCalTime $ pathToCalTime path) errors [])
           <?> "Helium logging"

--shouldnt this be a parser
pathToCalTime :: Path -> String
pathToCalTime (Path _ _ loggingdir _) = loggingdir -- stringToCalTime loggingdir

{-
   Parser for logging elements
-}
separator   :: Parser ()
separator   = skipMany1 (space <|> tab <|> char ',' <|> char ':' <?> "") 

pIdentifier    :: Parser String
pIdentifier = do c  <- letter
                 cs <- many (alphaNum <|> char '_')
                 return (c:cs)
              <?> "identifier"

pPhase :: Parser (Phase, [StaticErrorId])
pPhase = do p <- oneOf "LPRSTCI" 
            case p of
               'L' -> return (Lexical, [])
               'P' -> return (Parsing, [])
               'R' -> return (ResolOp, []) 
               'S' -> do errors <- sepBy pStaticError (oneOf ",;")
                         return (Static, errors)
                      <?> "static error code"
               'T' -> return (Typing, [])
               'C' -> return (CodeGen, [])
               'I' -> return (InternalErr, [])
         <?> "phase description"

pStaticError    :: Parser String
pStaticError = do many (alphaNum <|> char '_' <|> char '?')
               <?> "static error description"

pVersion :: Parser String
pVersion = do main <- many digit 
              sub <- many pSubversion
              return (main ++ (concat sub) )
           <?> "version"

pSubversion :: Parser String
pSubversion = do p <- char '.'
                 sub <- many alphaNum
                 return (p:sub)
              <?> "version element"

pBuildDate:: Parser String
pBuildDate = do char '('
                bd <- pDateString
                char ')'
                return bd
             <?> "Helium build date"

--pDateString :: Parser [String]
--pDateString = sepBy1 (many alphaNum) (oneOf ": ") <?> "date"

pDateString :: Parser String
pDateString = many (noneOf "()") 

{-
   Parser for logging path
-}
pPath :: Parser Path
pPath = do userdir <- pPathElement
           pPathSeperator
           loggingdir <- pPathElement
           pPathSeperator
           sourcefile <- pPathElement
           return (Path "" userdir loggingdir sourcefile)

pPathSeperator :: Parser ()
pPathSeperator = skipMany1 (char '/' <|> char '\\' <?> "")
           
pPathElement :: Parser String 
pPathElement = many (alphaNum <|> (oneOf "-@_.")) <?> "path description element"


toHelVer :: String -> String -> HeliumVersion
toHelVer version builddate = HeliumVersion version edited_bd --parsed_bd
 where edited_bd = map replaceSpecChar builddate          
                   where replaceSpecChar :: Char -> Char
                         replaceSpecChar c = case c of 
                              ':' -> '_'
                              ' ' -> '-'
                              otherwise  -> c
{-       parsed_bd = do calTime <- parse pDate "helium compiler build date" builddate
                      case (result) of
                           Left err  -> print err
                           Right xs  -> result <- xs
                      return result
-}

-- to parse a builddate nicely, not used yet
pDate :: Parser CalendarTime
pDate = do  weekday <- option Sunday pWeekDay
            option ' ' pDateSep
            month <- pMonth
            option ' ' pDateSep
            dayno <- pInt  
            pDateSep
            hour <- pInt
            pTimeSep
            min <- pInt
            pTimeSep
            sec <- pInt
            pDateSep
            timezone <- option "" pTimeZone   -- timezone is not allways provided
            pDateSep
            year <- pInt
            return (CalendarTime {ctYear = year
                                 ,ctMonth = month 
                                 ,ctDay = dayno 
                                 ,ctHour = hour
                                 ,ctMin = min 
                                 ,ctSec = sec 
                                 ,ctPicosec = 0 
                                 ,ctWDay = weekday 
                                 ,ctYDay = 0  -- not provided
                                 ,ctTZName = timezone
                                 ,ctTZ = 0  -- not provided
                                 ,ctIsDST = True
                                 }
                    )

pDateSep :: Parser Char
pDateSep = char ' '

pTimeSep :: Parser Char
pTimeSep = char ':'

pWeekDay :: Parser Day
pWeekDay = do string "Sun" <|> string "Sunday"
              return Sunday
           <|>
           do string "Mon" <|> string "Monday"
              return Monday
           <|>
           do string "Tue" <|> string "Tuesday"
              return Tuesday
           <|>
           do string "Wed" <|> string "Wednesday"
              return Wednesday
           <|>
           do string "Thu" <|> string "Thursday"
              return Thursday
           <|>
           do string "Fri" <|> string "Friday"
              return Friday
           <|>
           do string "Sat" <|> string "Saturday"
              return Saturday

pMonth :: Parser Month
pMonth = do string "Jan" <|> string "January"       
            return January
         <|>
         do string "Feb" <|> string "February"
            return February
         <|>
         do string "Mar" <|> string "March"
            return March
         <|>
         do string "Apr" <|> string "April"
            return April
         <|>
         do string "May" <|> string "May"
            return May
         <|>
         do string "Jun" <|> string "June"
            return June
         <|>
         do string "Jul" <|> string "July"
            return July
         <|>
         do string "Aug" <|> string "August"
            return August
         <|>
         do string "Sep" <|> string "September"
            return September
         <|>
         do string "Oct" <|> string "October"
            return October
         <|>
         do string "Nov" <|> string "November"
            return November
         <|>
         do string "Dec" <|> string "December"
            return December

pInt :: Parser Int
pInt = do number <- (many digit)
          return (read number::Int)

pTimeZone :: Parser String
pTimeZone = many alphaNum

-- unused
pEmptyLine :: Parser String
pEmptyLine = do many (oneOf "\t ")
                newline
                return ""	
             <?> "empty line"

word :: Parser String
word = many letter <?> "word"
