module Thesis.Helium.Logging where

----
-- Basic Logging datatypes
----

import Statistics.BasicTypes

import Thesis.Settings
import System.Time
import System.IO.Unsafe
import Data.List(product)
import Data.Char(ord)
import CommonIO
import Auxilary
import Data.Typeable  -- also added this to Phase in Logging.hs

data Logging = Log   {  username          ::   Username,
                        phase             ::   Phase,
                        heliumVersion     ::   HeliumVersion,
                        logPath           ::   Path,
                        logDate           ::   CalendarTime,
                        staticErrorCodes  ::   [StaticErrorId],
                        extraLoggingInfo  ::   ExtraLoggingInfo  -- extensions to the logging
                     }
    deriving (Show, Eq, Read, Typeable)

-- heliumErrorsFromCompiler :: [HeliumError],
-- sourceInput / compilerOutput

compareLogging l1 l2 = compare (logDate l1) (logDate l2)

instance Ord Logging where
   compare = compareLogging

-- Compilation phase, representing the phase in which the compilation process ended
-- Internal error is not an real phase, can be thrown in any phase
-- Specific static error code are defined separately
data Phase = Lexical | Parsing | ResolOp | Static | Typing | CodeGen | InternalErr
    deriving (Show, Eq, Read, Ord, Enum, Typeable)

data HeliumVersion =  HeliumVersion {   heliumVersionNumber :: String,
                                        buildDateRef :: String
                      }
   deriving (Eq, Read, Ord, Show, Typeable)

-- Log FilePath, including a base dir, simplifies the access to the program source of the logging
data Path   =  Path  { 
                  baseDir :: String,
                  userDir :: String,
                  loggingDir :: String,
                  sourceFile :: String
               }
   deriving (Eq, Read, Show, Typeable)

-- type CalendarDate = String -- CalendarTime  -- NewTimeLib Day -- defined in Auxilary.hs
type DataSet = [Logging]
type ExtraLoggingInfo = [(String, String)]

newtype Username = Username String
   deriving (Eq, Show, Read, Ord, Typeable)

type BuildDate = String
type StaticErrorId = String
type CompilerOutput = String


--------------------------------------------------------------------------------------------------------
--- Additional functions for reading loggings
--------------------------------------------------------------------------------------------------------

source  ::  Logging -> String
source  =   getSource

sourcecode              ::    Logging -> String
sourcecode              =     source

compileroutput          ::  Logging -> String
compileroutput          =   getCompilerOutput

compileroutputverbose   ::  Logging -> String
compileroutputverbose   =   getCompilerOutputVerbose

unparsedLogDate  ::  Logging -> String
unparsedLogDate  =   loggingDir . logPath

--  UnsafeperfomIO argument:
--  Getting the file contents is similar to querying a database (read only)
--  no changes to the world around, so unsafePerformIO can be used

getSource :: Logging -> String
getSource = unsafePerformIO . getContentsStrict . logInputPath

getCompilerOutput :: Logging -> String
getCompilerOutput = unsafePerformIO . getContentsStrict . logOutputPath

getCompilerOutputVerbose :: Logging -> String
getCompilerOutputVerbose = unsafePerformIO . getContentsStrict . logOutputPathVerbose

-- helper functions for forming the right paths
logInputPath :: Logging -> String
logInputPath = formatPath . getPathParts . logPath

logOutputPath :: Logging -> String
logOutputPath log =
   (logInputPath log) ++ outExt ++ (heliumVersionNumber helver) ++ "-" ++ ({-show $ -}buildDateRef helver)
   where
      helver = heliumVersion log
   
logOutputPathVerbose :: Logging -> String 
--logOutputPathVerbose logging@(Log _ _ (HeliumVersion version builddate) _ _) = 
--   (logInputPath logging) ++ outExtVerbose ++ version ++ "-" ++ builddate
logOutputPathVerbose log =
   (logInputPath log) ++ outExtVerbose ++ (heliumVersionNumber helver) ++ "-" ++ ({-show $ -}buildDateRef helver)
   where
      helver = heliumVersion log

getPathParts :: Path -> [String]
getPathParts (Path basedir userdir loggingdir sourcefile) = [basedir, userdir, loggingdir, sourcefile]

----
-- Instances for DataInfo
----

-- Settings for formating in and outputting date /times in gnuplot, lookat gnuplot> help date
loggingtimeInFmt :: String
loggingtimeInFmt = "%Y-%m-%d@%H_%M_%S"  -- from logfile: 2003-02-13@11_13_15_020

loggingtimeOutFmt :: String
loggingtimeOutFmt = "%Y-%m-%d\n%H:%M" -- "%Y-%W"  -- 2005 15


-- duplicate decl also in Presentation.hs
type LayoutFormat = String

-- helper functions for formulating the settings
formDataInputLayout :: Axis -> LayoutFormat -> String
formDataInputLayout axis format = "set " ++ show axis ++ "data time\n" ++ "set timefmt " ++ show format

formDataOutputLayout :: Axis -> LayoutFormat -> String
formDataOutputLayout axis format = "set format " ++ show axis ++ " " ++ show format

instance DataInfo Logging where
   dataKind _ = Interval
   showType _ =  "logging"
   shortname = showType
   ordIndex = error "no ordIndex definition for Logging" -- on date

instance DataInfo Phase where
   dataKind _ = Ordinal
   showType _ =  "phase"
   shortname = showType
   ordIndex = fromEnum
   allvalues = SummableBounded [Lexical, Parsing, ResolOp, Static, Typing, CodeGen, InternalErr]

instance DataInfo Username where
   dataKind _ = Nominal
   showType _ =  "username"
   ordIndex (Username name) = product $ map ord name -- some better ordering still to come
   shortname = showType

instance DataInfo HeliumVersion where
   dataKind _ = Nominal
   showType _ =  "helium version"
   shortname = showType
   allvalues = NoComplement

-- test data for Logging datatype
basedir = "/users/pkeeken/scratch/data/fp/archives/"

log0 = Log        (Username "peter")
                  Lexical
                  (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003")
                  (Path (basedir ++ "helium0203") "group62" "2003-03-13@11_13_15_020" "Hello.hs")
                  (stringToCalTime "2003-03-13@11_13_15_020")
                  []
                  []

log1 = Log {
                  username = Username "peter",
                  phase = Lexical,
                  heliumVersion = HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003",
                  logPath = Path (basedir ++ "helium0203") "group62" "2003-03-13@11_13_15_020" "Hello.hs",
                  logDate = (stringToCalTime "2003-03-13@11_13_15_020"),
                  staticErrorCodes = [],
                  extraLoggingInfo = []
                }
log2 = Log {
                  username = Username "peter",
                  phase = Parsing,
                  heliumVersion = HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003",
                  logPath = Path (basedir ++ "helium0203") "group29" "2003-03-13@11_25_00_599" "Hello.hs",
                  logDate = (stringToCalTime "2003-03-13@11_25_00_599"),
                  staticErrorCodes = [],
                  extraLoggingInfo = []
                }

log2b_3 = Log {
                  username = Username "group84",
                  phase = Parsing,
                  heliumVersion = HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003",
                  logPath = Path (basedir ++ "helium0203") "group29" "2003-03-13@11_25_00_599" "Hello.hs",
                  logDate = (stringToCalTime "2003-03-13@11_25_00_599"),
                  staticErrorCodes = [],
                  extraLoggingInfo = []
                }
                
-- imports should be available (use calling of main, instead of compile in Compile
log3 = Log {
                  username = Username "group84",
                  phase = ResolOp,
                  heliumVersion = HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003", --1.1A
                  logPath = Path (basedir ++ "helium0203") "group84" "2003-04-17@07_49_00_357" "HTML.hs",
                  logDate = (stringToCalTime "2003-04-17@07_49_00_357"),
                  staticErrorCodes = [],
                  extraLoggingInfo = []
                }

log4 = Log {
                  username = Username "group38",
                  phase = Static {-[""]-},
                  heliumVersion = HeliumVersion "1.1A" "Mon-Mar-10-09_19_12--2003", --1.1A
                  logPath = Path (basedir ++ "helium0203") "group38" "2003-04-02@11_24_37_536" "Log.hs",
                  logDate = (stringToCalTime "2003-04-02@11_24_37_536"),
                  staticErrorCodes = [],
                  extraLoggingInfo = []
                }


log5 = Log {
                  username = Username "peter",
                  phase = Typing,
                  heliumVersion = HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003",
                  logPath = Path (basedir ++ "helium0203") "group86" "2003-03-13@14_53_58_103" "Opgaven.hs",
                  logDate = (stringToCalTime "2003-03-13@14_53_58_103"),
                  staticErrorCodes = [],
                  extraLoggingInfo = []
                }

log6 = Log {
                  username = Username "group38",
                  phase = Typing {-[""]-},
                  heliumVersion = HeliumVersion "1.1A" "Mon-Mar-10-09_19_12--2003", --1.1A
                  logPath = Path (basedir ++ "helium0203") "group38" "2003-04-02@11_24_37_536" "Log.hs",
                  logDate = (stringToCalTime "2003-04-02@11_24_37_536"),
                  staticErrorCodes = [],
                  extraLoggingInfo = []
                }

-- for testing purposes
logs :: [Logging]
logs = [log1, log2, log3, log4, log5,log6]

{-
type SourceInput = IORef (IO String)

2 options for reading the input
--as io action (in comb with a IORef to for come multiple reading of the same logging
safe: do {x <- newIORef (do { x <- getCO path; return (x, False) }); return (Log ... x)}

readSource :: IORef (IO a) -> IO a
readSource source = do
        action          <- readIORef source
        (s,cached)      <- action
        when (!cached)  (writeIORef source (return (s, True)))
        return s

-- forced read, unsafe, but the input is regarded unchanged so can be allowed
unsafe: Log ... (unsafePerformIO (getCOPath))
-}

{-
   For accessing source and compiler output

-- /scratch/data/fp/archives/helium0203/group86/2003-02-13@14_53_58_103/Opgaven.hs
--------------------------------------------------------------------------
-- Test Data
--------------------------------------------------------------------------
tstLoggingFP0203, tstLoggingFP0304 :: Logging
tstLoggingFP0203 = Log "group86" Typing (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003") (Path "/scratch/data/fp/archives/helium0203/" "group86" "2003-02-13@14_53_58_103" "Opgaven.hs") "2003-02-13@14_53_58_103" [] --he0203 logging
tstLoggingFP0304 = Log "group80" Lexical (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "/scratch/data/fp/archives/fp-0304-p3/" "group80" "2004-01-03@12_46_12_871" "Hello.hs") "2003-02-13@14_53_58_103" [] -- fp0304 logging
-}
--tstLoggingsFP0304 :: [Logging]
--tstLoggingsFP0304 = [Log "group80" Lexical (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group80" "2004-01-03@12_46_12_871" "Hello.hs") "2004-01-03@12_46_12_871",Log "group91" Lexical (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group91" "2004-01-03@12_47_19_110" "Complex.hs") "2004-01-03@12_47_19_110",Log "group91" (Static [""]) (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group91" "2004-01-03@12_48_13_116" "Complex.hs") "2004-01-03@12_48_13_116",Log "group80" CodeGen (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group80" "2004-01-03@12_48_16_948" "Hello.hs" "2004-01-03@12_48_16_948"),Log "group91" (Static [""]) (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group91" "2004-01-03@12_48_51_640" "Complex.hs"),Log "group80" CodeGen (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group80" "2004-01-03@12_49_03_745" "Hello.hs"),Log "group91" Typing (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group91" "2004-01-03@12_49_24_165" "Complex.hs"),Log "group81" CodeGen (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group81" "2004-01-03@12_49_47_359" "Pretty.hs"),Log "group81" Lexical (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group81" "2004-01-03@12_52_28_407" "Hello.hs"),Log "group81" CodeGen (HeliumVersion "1.2" "Tue-Jan-27-14_56_04-WEST-2004") (Path "group81" "2004-01-03@12_53_14_514" "Hello.hs")]

--tstLoggingsHe0203 :: [Logging]
--tstLoggingsHe0203 = [Log "group86" Typing (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003") (Path "group86" "2003-02-13@14_53_58_103" "Opgaven.hs"),Log "group86" Typing (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003") (Path "group86" "2003-02-13@14_54_28_742" "Opgaven.hs"),Log "group86" Typing (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003") (Path "group86" "2003-02-13@14_54_58_172" "Opgaven.hs"),Log "group86" Typing (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003") (Path "group86" "2003-02-13@14_55_10_993" "Opgaven.hs"),Log "group86" Typing (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003") (Path "group86" "2003-02-13@14_55_34_272" "Opgaven.hs"),Log "group86" Typing (HeliumVersion "1.1" "Mon-Mar-10-09_19_12--2003") (Path "group86" "2003-02-13@17_14_04_865" "Opgaven.hs")]

tstOutput :: String
tstOutput = "====================================================\n= Helium Compiler 1.2-Tue-Jan-27-14_56_04-WEST-2004\n====================================================\nCompiling group80/2004-01-03@12_46_12_871/Hello.hs\n (1,31): Non-terminated character literal\n   Hint: Correct examples of Chars: 'a' '\\n' '&'\n         To write a function in infix notation, use backquotes: `hello`\n (1,31): Non-terminated character literal\n   Hint: Correct examples of Chars: 'a' '\\n' '&'\n         To write a function in infix notation, use backquotes: `hello`\n Compilation failed with 2 error\n"

tstOutput2:: String
tstOutput2 = "====================================================\n= Helium Compiler 1.1-Mon-Mar-10-09_19_12--2003\n====================================================\nCompiling group86/2003-02-13@14_53_58_103/Opgaven.hs\n (20,33): Warning: in variable\n  expression       : ++\n    type           : [a] -> [a] -> [a  ]\n    expected type  : Int -> b   -> [Int]\n  probable fix     : use : instead\n \n (20,38): Type error in infix application\n  expression       : b ++ bs\n  operator         : ++\n    type           : [a] -> [a  ] -> [a]\n    does not match : Int -> [Int] -> b  \n \n Compilation failed with 2 errors\n"
