module Thesis.HeliumUniversal.LexerToken where

-- also lookat: http://www.haskell.org/onlinereport/syntax-iso.html

import Thesis.HeliumUniversal.BasicTypes
import Data.Typeable

data TokenAnalysisResult = TokenAnalysisResult {
                                ppTokensWithLayout :: String,
                                tokens :: [Token],
                                tokenWithLayout :: [Token]
                             }
          deriving (Show,Read, Eq)

-- LexerTokens similar for all helium versions, (custom show only differs)
type Token      = (SourcePos,Lexeme)

data Lexeme =
   -- literals
      LexChar            String
    | LexString          String
    | LexInt             String
    | LexFloat           String

   
    | LexVar             String
    | LexVarSym          String
    | LexCon             String
    | LexConSym          String

   --  reserved words and (special) symbols
    | LexKeyword         String  -- see keywords
    | LexResVarSym       String  -- see reservedVarSym
    | LexResConSym       String  -- see reservedConSyms
    | LexSpecial         Char

   -- extra layout
    | LexInsertedOpenBrace  -- { inserted because of layout
    | LexInsertedCloseBrace -- }
    | LexInsertedSemicolon  -- ;

    | LexEOF
    deriving (Eq, Show, Read)

-- show similar for all helium, except for last line
instance HeliumCustomShow Lexeme where
    heliumCustomShow x = case x of 
        LexChar    c        -> "character literal"      ++ " '" ++ c      ++ "'"
        LexString  s        -> "string literal"         ++ " \""++ s      ++ "\""
        LexInt     i        -> "integer literal"        ++ " '" ++ i      ++ "'"
        LexFloat   f        -> "floating-point literal" ++ " '" ++ f      ++ "'"

        LexVar     n        -> "variable"               ++ " '" ++ n      ++ "'"
        LexVarSym  o        -> "operator"               ++ " '" ++ o      ++ "'"
        LexCon     c        -> "constructor"            ++ " '" ++ c      ++ "'"
        LexConSym  o        -> "constructor operator"   ++ " '" ++ o      ++ "'"
        
        LexKeyword kwd      -> "keyword '" ++ kwd ++ "'"
        LexResVarSym s      -> "'" ++ s ++ "'"
        LexResConSym s      -> "'" ++ s ++ "'"
        LexSpecial c        -> "'" ++ [c] ++ "'"
        
        LexInsertedOpenBrace  -> "inserted '{'"
        LexInsertedCloseBrace -> "end of block (based on layout)"
        LexInsertedSemicolon  -> "next in block (based on layout)"
                        
        LexEOF              -> "end of file"
        
        _                   -> "unknown"   -- deleted since helium 1.1.1


-----------------------------------------------------------
-- Constants, Most General, Small differences are encoded in subsequent hierar sourcefiles
-----------------------------------------------------------

escapeChars :: String
escapeChars = "\\nabfnrtv\"'"

symbols :: String
symbols = "!#$%&*+./<=>?@^|-~:\\"

keywords :: [String]  -- pvk become LexKeyword
keywords = 
    [ "let", "in", "do", "where", "case", "of", "if"
    , "then", "else", "data", "type", "module", "import", "hiding"  -- pvk "hiding" since helium 1.5
    , "infix", "infixl", "infixr", "_", "deriving"
    , "class", "instance", "default", "newtype" -- not supported
    ]

reservedConSyms :: [String]  -- pvk become LexResConSym
reservedConSyms =
    [ "::" ]

reservedVarSyms :: [String]  -- pvk become LexResVarSym
reservedVarSyms =
    [ "=>", "->", "<-", "..", "-", "-.", "@", "=", "\\", "|", "~" ]  -- pvk "~" since helium 1.5

specialsWithoutBrackets :: String
specialsWithoutBrackets = 
    ",`;"

reserveStrategyNames :: [Token] -> [Token]
reserveStrategyNames tokens = 
  map (\token@(pos, lexeme) -> case lexeme of 
                   LexVar s    | s `elem` strategiesKeywords -> (pos, LexKeyword s)
                   LexVarSym s | s == "=="                   -> (pos, LexResVarSym s)
                   LexConSym s | s == ":"                    -> (pos, LexResConSym s)
                   _ -> token
      ) 
      tokens

-- pvk added since helium 1.1.2
strategiesKeywords :: [String]
strategiesKeywords = [ "phase", "constraints", "siblings" ]

falsekeywordsHelium1v1Helium1v1v1 :: [String] 
falsekeywordsHelium1v1Helium1v1v1 = [ "phase", "constraints"]

-- to standardize over different tokenized representations of a source
toUniTokensLaterThenHelium1v1v2 :: [Token] -> [Token]
toUniTokensLaterThenHelium1v1v2 tokens =
     map (\token@(pos, lexeme) -> case lexeme of 
                   LexKeyword s| s `elem` falsekeywordsHelium1v1Helium1v1v1
                                                               -> (pos, LexVar s)
                   LexResVarSym s | s == "=="                  -> (pos, LexVarSym s)
                   LexResConSym s | s == ":"                   -> (pos, LexConSym  s)
                   
                   _                                           -> token
      ) 
      tokens

toUniTokensLaterThenHelium1v5 :: [Token] -> [Token]
toUniTokensLaterThenHelium1v5 tokens = toUniTokensLaterThenHelium1v1v2 $
     -- compensate for new keyword "hiding"? isnt supported in new version
     -- compensate for "~" in symbols list
{-
     map (\token@(pos, lexeme) -> case lexeme of 
                   LexKeyword s | s == "hiding"                  -> (pos, LexString s)
                   
                   _                                           -> token
      )
-}
      tokens

-- lexemes categorised 
anyToken :: [Token]
anyToken = undefined

literalTokens :: [Lexeme]
literalTokens = [
    LexChar            ""
  , LexString          ""
  , LexInt             ""
  , LexFloat           ""
  ]

languageConstructs :: [Lexeme]
languageConstructs = [
     LexVar             ""
   , LexVarSym          ""
   , LexCon             ""
   , LexConSym          ""
   ]

reservedWords = [
     LexKeyword         ""  -- see keywords
   , LexResVarSym       ""  -- see reservedVarSym
   , LexResConSym       ""  -- see reservedConSyms
   , LexSpecial         ' '
   ]

insertedLayout = [
     LexInsertedOpenBrace  -- { inserted because of layout
   , LexInsertedCloseBrace -- }
   , LexInsertedSemicolon  -- ;
   ]
