module Thesis.Main where

import Statistics.DescriptiveAnalysis
import Statistics.Presentation

import Thesis.AnalysisPredefined
import Thesis.Helium.LoggingPP
import Thesis.Helium.LogfileParser
import Thesis.Helium.Logging
import Thesis.Args
import Thesis.Select
import Thesis.Settings
import Thesis.MainI
import Auxilary

import Data.Maybe
import Data.List
import Control.Monad
import System.Time
import System.Exit
import System.Environment -- (withArgs)
import System.IO.Unsafe(unsafePerformIO)
import Directory
import qualified Control.Exception

--------------------------------------
-- Run Analysis Tool Set
--------------------------------------

main :: IO ()
main =
   do
      args <- getArgs
      (options, fullName) <- processArgs args  -- does also output dir check
      
      -- Check if Logfile must exist, this test doesn't use the search path
      fileExists <- doesFileExist fullName
      when (not fileExists) $ do
         putStrLn $ "Can't find logfile " ++ show fullName
         exitWith (ExitFailure 1)

      if (not $ isOutputFilePathOption `elemBy` options)
         then 
            putStrLn $ "No output directory specified, output written to stdout."
         else
            return ()

      -- extract outputpath from options
      let outfilepath   = ((getValIfOption isOutputFilePathOption options) ++ dirsep )

      -- extract outputformat from options
      let opt = find isOutputFormatOption options
      outfileformat <- if (isJust $ opt)
                             then do let (OutputFormat fmt) = fromJust opt
                                     parseFmtOpt fmt   -- throws warning if not recognized
                             else return outputformat_default
      when debug $ putStrLn $ "Outputformat tmp: " ++ show outfileformat

      putStrLn "Parsing logfile..."
      logs <- parseLogfile fullName

      if (RunAllAnalyses `elem` options)
         then
         -- run all analysis
         do putStrLn "Running all predefined analyses..."
            mapM (\research -> research outfileformat outfilepath (kickStartAnalysis logs startkey)) $
                map snd predefAnalyses
            -- return ()
         else
            -- run specified analyses
            do let anas = getPredefinedAnalysis options
               when (length anas == 0) $ do
                  putStrLn "No analysis specified in options"
                  exitWith (ExitFailure 1)
               mapM (runAnalyses options logs outfileformat) anas

      return ()

runAnalyses :: [Option] -> [Logging] -> OutputFormat -> Int -> IO ()
runAnalyses options logs format num =
      if (num < length predefAnalyses)
         then
            Control.Exception.catch
               (
               -- run Analysis
               do -- let fstLine = "# Analysis " ++ show num ++ ": " ++ getPreDefAnalysisID num predefAnalyses
                  let fstLine = "#" ++ (fst $ (!!) predefAnalyses num)
                      ana = snd $ (!!) predefAnalyses num
                  if (isOutputFilePathOption `elemBy` options)
                     then do
                        -- run analysis with result written to outputpath
                        let outfilepath   = ((getValIfOption isOutputFilePathOption options) ++ dirsep )
                        -- anaResult <- runPreDefAnalysis (Just outfilepath) num predefAnalyses logs
                        anaresult <- ana format outfilepath (kickStartAnalysis logs startkey)
                        putStrLn $ fstLine ++ " written to " ++ outfilepath
                     else do
                        -- run analysis with result written to stdout
                        -- only pure presentations are printed on screen
                        -- throw warning??
                        -- anaResult <- runPreDefAnalysis Nothing num predefAnalyses logs
                        putStrLn $ fstLine
                        -- putStrLn $ anaResult
                        
                  return ()
               )
               -- error message if an error occures
               (\e -> do
                         -- putStrLn $ "Error occured while running analysis " ++ show num ++ ": " ++ getPreDefAnalysisID num predefAnalyses
                         putStrLn $ "Error occured while running analysis: " ++ "#" ++ (fst $ (!!) predefAnalyses num) -- fstLine
                         putStrLn $ "Error: " ++ show e
                         putStrLn $ "-- Analysis skipped --"
                         return ()
               )
         else
            do
               putStrLn $ "Analysis " ++ show num ++ " is not defined. Analysis skipped"
               --exitWith (ExitFailure 1)


{-
runAnalyses :: [Option] -> [Logging] -> Int -> IO ()
runAnalyses options logs num = 
   if (num < length predefAnalyses)
         then
            Control.Exception.catch
               (
               -- run Analysis
               do let fstLine = "# Analysis " ++ show num ++ ": " ++ getPreDefAnalysisID num predefAnalyses
                  if (isOutputFilePathOption `elemBy` options)
                     then do
                        let outfilepath   = ((getValIfOption isOutputFilePathOption options) ++ dirsep )
                        anaResult <- runPreDefAnalysis (Just outfilepath) num predefAnalyses logs
                        putStrLn $ fstLine ++ " written to " ++ outfilepath
                     else do  -- only pure presentations are printed on screen 
                        anaResult <- runPreDefAnalysis Nothing num predefAnalyses logs
                        putStrLn $ fstLine
                        putStrLn $ anaResult
                  return ()
               )
               -- error message if an error occures
               (\e -> do
                         putStrLn $ "Error occured while running analysis " ++ show num ++ ": " ++ getPreDefAnalysisID num predefAnalyses
                         putStrLn $ "Error: " ++ show e
                         putStrLn $ "-- Analysis skipped --"
                         return ()
               )
         else
            do
               putStrLn $ "Analysis " ++ show num ++ " is not defined. Analysis skipped"
               --exitWith (ExitFailure 1)

-- running of the old analysis type
runPredefAnalysis options logs num =
   if (num < length predefAnalysesOld)
         then 
            do let fstLine = "# Analysis " ++ show num ++ ": "
               if (isOutputFilePathOption `elemBy` options)
                  then do
                     let outfilepath   = ((getValIfOption isOutputFilePathOption options) ++ dirsep ++ "analysis" ++ show num)
                         anaResult     = runPreDefAnalysisType outfilepath num predefAnalysesOld logs

                     -- write analysis result to file
                     writeFile
                        (outfilepath ++ ".out")
                        (fstLine ++ "\n" ++ anaResult)
                     putStrLn $ fstLine ++ " written to " ++ outfilepath
                  else do  -- mind side effect of gnuplot generation
                     let anaResult = runPreDefAnalysisType "no-output" num predefAnalysesOld logs
                     putStrLn $ fstLine
                     putStrLn $ anaResult
               return ()
         else
            do
               putStrLn $ "Analysis " ++ show num ++ " is not defined. Analysis skipped"
               --exitWith (ExitFailure 1)
-}
testloggings1  = "/scratch/data/fp/archives/helium0203/fp0203_recompilation_onlyCompilesTst1000.log"
testloggings   = "/scratch/data/fp/archives/helium0203/fp0203_recompilation_onlyCompilesTst100.log"
testloggings2  = "/scratch/data/fp/archives/helium0203/fp0203_recompilation_onlyCompilesTstRandomSelection.log"

testArgs = ["-A" {-, "1"-}, "-o", testoutputdir, testloggings1]

testoutputdir = "/users/pkeeken/scratch/tmp/analyses_output/test_results_27_3_06/"

testMain = withArgs testArgs main
