module Thesis.Research.HintRateResearch where

-- next to should be combined, in single default AC import
import Statistics.DescriptiveAnalysis
import Statistics.SummaryMeasures
import Statistics.Presentation

import Auxilary
import Thesis.Helium.Logging
import Thesis.Metrics
import Thesis.Research.Common

import Text.ParserCombinators.Parsec
import Text.Regex

import System.IO.Unsafe (unsafePerformIO)
{-

----
-- Hint Ratio Analysis
----
calculates the percentage of Hints overall to the type errors

approach:
  filter on typing phase loggings
  count number of errors
  count number of hints per error

-- useful: cat fp0405_original.log | grep "\:T:" | cut -d: -f7 | cut -d\/ -f1,2 | xargs -i find {} -type f -iname "*out*" | xargs -i cat {} | m

or better:
grep ":T:" fp0405_originalFst1000.log | cut -d: -f7 | xargs -i bash -c 'cat {}\.out* | grep -c probable\ fix'  |  grep -v 0 | wc -l
112
grep ":T:" fp0405_originalFst1000.log | cut -d: -f7 | xargs -i bash -c 'cat {}\.out* | grep probable\ fix'  | grep -v 0 | wc -l
165

 grep ":T:" fp0405_originalFst1000.log | cut -d: -f7 | xargs -i bash -c 'cat {}\.out* | grep -c probable\ fix' > aantalAll.txt

pb: (fp 0405)
1000 -> 112
2000 -> 199
all fp0405 -> 915

-}

data Percentage = Percentage Double Int Int

instance Show Percentage where
   show (Percentage perc x total) = show perc ++ "% (" ++ show x ++ " / " ++ show total ++ ")"

default_logfile    = "datasets/demo/demo.log"
default_outputpath = "analysis_results/demo"

typeHintsAnalysis :: AnalysisFK KeyHistory Logging [String]
typeHintsAnalysis =
   basicAnalysis "type hint(s)" (concat . multipleMatchRegex hint_regex . compileroutput)
   where
      hint_regex = mkRegexWithOpts " *(probable fix.*)" True True

countNumberOfLoggingsWithHint :: AnalysisFK KeyHistory [Logging] Int
countNumberOfLoggingsWithHint =
               basicAnalysis "number of compiles with hint" (length . filter (not . null))
        <.>    mapAnalysis typeHintsAnalysis 

numberOfTypeHintsAnalysis :: AnalysisFK KeyHistory [Logging] (Int, Int)
numberOfTypeHintsAnalysis =
         (countNumberOfLoggings <&> countNumberOfLoggingsWithHint)
   <.>   basicAnalysis  "(from only type incorrect programs)" (filter ((== Typing) . phase))

-- used by TypingTillCodeResearch
percentageOfTypeHintsAnalysis'  ::  AnalysisFK KeyHistory [Logging] Double
percentageOfTypeHintsAnalysis'  =
         basicAnalysis "percentage" (uncurry (flip percentage))
   <.>   numberOfTypeHintsAnalysis

percentageOfTypeHintsAnalysis :: AnalysisFK KeyHistory [Logging] Percentage
percentageOfTypeHintsAnalysis =
         basicAnalysis "percentage" (\(total, x) -> Percentage (percentage x total) x total)
   <.>   numberOfTypeHintsAnalysis

main = runNewResearchFromLogFile hintRatioResearch startkey default_logfile default_outputpath HtmlFormat

hintRatioResearch :: OutputFormat -> [(KeyHistory, [Logging])] -> FilePath -> IO ()
hintRatioResearch format input outputpath = do
   let
      name = "Hint Ratio Research"
      researchdir = joinPaths outputpath (stringToFilePath name)
      outputfile  = joinDirFileExt researchdir (stringToFilePath name) $ showFormatExt format

   createDirectoryIfMissing True researchdir

   barChartPerc <- renderBarChart researchdir $ percentageOfTypeHintsAnalysis' <.> groupPerWeek <$> input

   -- putStrLn $ show $ showAsTable1D $ numberOfTypeHintsAnalysis input

   writeFile outputfile
                  $ renderMarkUpDoc format $ document noHtml $
                    (section Nothing ! [formatonly $ show HtmlFormat] << name)
                      +++
                   (showAsTable1D $ percentageOfTypeHintsAnalysis <$> input)
                     +++
                   (showAsTable1D $ percentageOfTypeHintsAnalysis <.> groupPerWeek <$> input)
                     +++
                   plotToFigure' researchdir barChartPerc

   return ()


{-

-- More specific output parser


-- parseOutput = basicAnalysis "compileroutput" (parse parseHeliumOutput "")

data HeliumOutputVerbose = HeliumOutputVerbose {
      sourcefile :: FilePath,
      lexMsg :: String,
      parseMsg :: String,
      resolvOpMsg :: String,
      staticMsg :: String,
      typingMsg :: String,
      endMsg :: String
   }
   deriving (Show, Eq, Read)

parseHeliumOutput' :: String -> HeliumOutputVerbose
parseHeliumOutput' source =
   let output = filter ((==[' ']) . take 1) $ lines source
       filepath = break (==" Lexing...") output
       lexlines = break (==" Parsing...") output
       parselines = break (==" Importing...") $ snd lexlines
       importlines = break (==" Resolving operators...") $ snd parselines
       resoplines = break (==" Static checking...") $ snd importlines
       staticlines = break (==" Type inference directives...") $ snd resoplines
       typedirlines = break (==" Type inferencing...") $ snd staticlines
       typeinflines = break (==" Desugaring...") $ snd typedirlines
       desuglines = break (==" Code generation...")  $ snd typeinflines
       codegenlines = break (==" balabl...") $ snd desuglines
   in
      HeliumOutputVerbose
         ""
         (unlines $ drop 1 $ fst lexlines)
         (unlines $ drop 1 $ fst $ parselines)
         (unlines $ drop 1 $ fst $ resoplines)
         (unlines $ drop 1 $ fst $ staticlines)
         (unlines $ drop 1 $ fst $ typeinflines)
         (unlines $ drop 1 $ fst $ desuglines)

test = parseHeliumOutput' exampleoutput1

exampleoutput1 = "====================================================\n= Helium Compiler 1.5.1-Tue-Mar--8-15_53_21-RST-2005\n====================================================\nCompiling group62/2005-03-07@11_08_19_216/Proplogic.hs\n Lexing...\n Parsing...\n parselinges\n Importing...\n Resolving operators...\n Static checking...\n Type inference directives...\n Type inferencing...\n (62,1): Warning: Missing patterns in function bindings:\n   variabelen (En []) = ...\n   variabelen (Of []) = ...\n (82,1): Warning: Missing patterns in function bindings:\n   toon (En []) = ...\n   toon (Of []) = ...\n Desugaring...\n Code generation...\n Compilation successful with 2 warnings \n"

exampleoutput2 ="Compiling group3/2005-01-07@16_13_39_137/CA1.hs\n Lexing...\n Parsing...\n Importing...\n Resolving operators...\n Static checking...\n Type inference directives...\n Type inferencing...\n   (58,13): Type error in variable\n   expression       : foldl\n     type           :(a  -> b   -> a  ) -> a       -> [b] -> a\n     expected type  : ([c] -> [c] -> [c]) -> [[Int]] -> [Int]\n  probable fix     : use foldl1 instead\n\n   Compilation failed with 1 error\n"


parseHeliumOutput :: Parser HeliumOutputVerbose -- HeliumTypeErrorParsed
parseHeliumOutput =
   do
      string "Compiling"
      file <- parsePath
      return $ HeliumOutputVerbose file "" "" "" "" "" ""
      <?> "Verbose Helium compiler output"

parsePath = do
   many anyChar

{-
getTypeHintsAnaT :: AnalysisFK KeyHistory Logging [(String, String, String, [String])]
getTypeHintsAnaT =
   basicAnalysis ""
      ({-concat
      .-} multipleMatchRegexAll (mkRegexWithOpts " *(probable fix.*)" True True)
      . typingMsg
      . parseHeliumOutput'
      . m1_1_3
      )-}

   -}
