module Thesis.Research.LineBasedInterpretationResearch where

-- next to should be combined, in single default AC import
import Statistics.DescriptiveAnalysis
import Statistics.Presentation
-- import Statistics.BasicTypes

import Thesis.Metrics -- (m1_2_1b)  -- same as linelexer
import Thesis.Helium.Parser.ResearchLexer

import Thesis.Research.Common
import Thesis.Research.AbstractIntMetricResearch
import Thesis.Helium.Logging
import Data.Typeable
import Control.Monad



import Auxilary
import Grouping

import Data.Maybe
import System.Time

{-

Analyses based in the line based interpretatino of a source file, using Thesis.Helium.Parser.ResearchLexer.linelexer :: String -> [SourceSegment]

Length of Module analysis or Lines of Code (LOC)

-}

default_logfile    = "datasets/demo/demo.log"
default_outputpath = "analysis_results/demo"

metricLOSourceSegments :: AnalysisFK KeyHistory Logging Int
metricLOSourceSegments = calculateNumber_OfElements
             <.> segmentMetric

metricLOC = calculateNumber_OfElements
           <.> basicAnalysis "Line of Code (LOC)"
                             (map segmentLines . filter isCodeLine . linelexer . sourcecode)

metricLOD = calculateNumber_OfElements
           <.> basicAnalysis "Line of Documentation (LOD)"
                             (map segmentLines . filter (\l -> isMultiCommentLine l || isSingleCommentLine l) . linelexer . sourcecode)

metricLOE = calculateNumber_OfElements
           <.> basicAnalysis "Empty lines (EL)"
                             (map segmentLines . filter (isEmptyLine) . linelexer . sourcecode)

-- miss niet zo zinnig
ratioAnalysis :: AnalysisFK KeyHistory (Int, Int) Double
ratioAnalysis = basicAnalysis "ratio" (\(v1, v2) -> (realToFrac v1 / realToFrac v2))

-- metricLOCLODRatio :: (Fractional a) => AnalysisFK KeyHistory Logging a
metricLOCLODRatio =
         ratioAnalysis
     <.> (metricLOD <&> metricLOC)

segmentMetric  ::    AnalysisFK KeyHistory Logging [SourceSegment]
segmentMetric  =     basicAnalysis "Line of Source" (segmentlexer . sourcecode)

lineSegmentMetric    ::    AnalysisFK KeyHistory Logging [SegmentType]
lineSegmentMetric    =
   basicAnalysis "Line of Source" (concatMap flattenSegments . segmentlexer . sourcecode)

-- simplified data for analysis
data SegmentType = CodeLn | CommentLn | EmptyLn
   deriving (Show, Eq, Ord, Enum)

flattenSegments :: SourceSegment -> [SegmentType]
flattenSegments segment@(CodeLine _)           =  replicate (length $ segmentLines segment) CodeLn
flattenSegments segment@(SingleCommentLine _)  =  [CommentLn]
flattenSegments segment@(MultiCommentLine _)   =  replicate (length $ segmentLines segment) CommentLn
flattenSegments segment@(EmptyLine _)          =  replicate (length $ segmentLines segment) EmptyLn

countSourceSegmentsMetric  ::  AnalysisFK KeyHistory Logging [(SegmentType, Int)]
countSourceSegmentsMetric  =
         basicAnalysis "" countOrdValues
   <.>   lineSegmentMetric

countOrdValues  ::  Ord a => [a] -> [(a,Int)]
countOrdValues  =   annotateList length . groupAll

annotateList :: ([a] -> b) -> [[a]] -> [(a, b)]
annotateList fun = map (\list -> (head list, fun list))

segmentDistribution :: ([Int] -> b) -> AnalysisFK KeyHistory [Logging] [(SegmentType, b)]
segmentDistribution aggregatefun    =
         basicAnalysis "" (map aggregateList . (groupAllUnder fst) . concat)
   <.>   mapAnalysis countSourceSegmentsMetric
 where
    aggregateList list = (fst $ head list, aggregatefun $ map snd list)

medianSegmentDistribution = segmentDistribution median

relmedianSourceLineDistribution :: AnalysisFK KeyHistory [Logging] [(SegmentType, Double)]
relmedianSourceLineDistribution =
         relativeValueCount
   <.>   medianSegmentDistribution

-- combined rel and abs, also in PhaseResearch
data AbsRelValue a = AbsRel a Float

instance Show a => Show (AbsRelValue a) where
   show (AbsRel abs rel) = show abs ++ " (" ++ show rel ++ "%)"

{-
-- necessary for ordValueCount
instance DataInfo SegmentType where
   dataKind v = Ordinal
   shortname v = "Segment type"
   ordIndex = fromEnum
-}

-- for table with combined abs/rel values
-- absRelAnalysis :: AnalysisFK KeyHistory [Logging] [(SegmentType, AbsRelValue)]
absRelAnalysis =
       (basicAnalysis "" (map (\(absT,relT) -> (fst absT, AbsRel (snd absT) (snd relT))) . uncurry zip))
                        <.>  ((basicAnalysis "" id) <&> relativeValueCount)
                        -- <.>  ordValueCount  -- [(a,Int)]
                        <.>  medianSegmentDistribution  -- [(SegmentType, b)]
                        <.>  groupPerWeek

main = runNewResearchFromLogFile' linebasedResearchOverall startkey default_logfile default_outputpath HtmlFormat
mainPerStudent =
   runNewResearchFromLogFile' linebasedResearchPerStudent startkey default_logfile default_outputpath HtmlFormat

researchname = "Line Based Module Interpretation Research"

linebasedResearchOverall :: OutputFormat -> FilePath -> [(KeyHistory, [Logging])] -> IO ()
linebasedResearchOverall format outputpath input = do
   let
      researchref = stringToFilePath researchname
      researchdir = outputpath ++ "/" ++ researchref  ++ "/"
      outputfile = researchdir ++ researchref  ++ "." ++ (showFormatExt format)

   createDirectoryIfMissing True researchdir

   research <- linebasedResearch [] format researchdir input
   
   writeFile outputfile
             (renderMarkUpDoc format $ document noHtml $ research)
   return ()

linebasedResearchPerStudent :: OutputFormat -> FilePath -> [(KeyHistory, [Logging])] -> IO ()
linebasedResearchPerStudent format outputpath input = do
   let
      researchref = stringToFilePath (researchname ++ "perstudent")
      researchdir = outputpath ++ "/" ++ researchref  ++ "/"
      outputfile = researchdir ++ researchref  ++ "." ++ (showFormatExt format)

   createDirectoryIfMissing True researchdir

   researchPerStudent <-
      researchPerDataSet linebasedResearch "student" format researchdir ({-take 10 $ -}groupPerStudent <$> input)

   writeFile outputfile
             (renderMarkUpDoc format $ document noHtml $
                  concatMarkUp $ researchPerStudent)

   return ()


linebasedResearch :: String -> OutputFormat -> FilePath -> [(KeyHistory, [Logging])] -> IO MarkUpDoc
linebasedResearch partname format outputpath input = do
   let
      -- name = "Line Based Module Interpretation Research"
      researchdir = joinPaths outputpath (stringToFilePath partname)
      -- outputfile  = joinDirFileExt researchdir (stringToFilePath name) $ showFormatExt format

   createDirectoryIfMissing True researchdir

--    locResearchLOC <- abstractIntMetricResearchInTime metricLOC (partname ++ "LOC") format input outputpath
--    locResearchLOD <- abstractIntMetricResearchInTime metricLOD (partname ++ "LOD") format input outputpath
--    locResearchLOE <- abstractIntMetricResearchInTime metricLOE (partname ++ "EL") format input outputpath

--  miss niet zo zinnig
--    locResearchLOD_LOD_Ratio <-
--          abstractIntMetricResearchInTime metricLOCLODRatio "LOD_LOC ratio" format input researchdir

   abslinedistribution <- renderStackedBarChart researchdir  $ medianSegmentDistribution
                                                       <.> groupPerWeek <$> input

   -- by rearranging the analysis, renderRelativeStackedBarChart could also be used
   rellinedistribution <- renderStackedBarChart researchdir  $
                                                           relmedianSourceLineDistribution
                                                       <.> groupPerWeek <$> input

   when (format == LaTeXFormat || format == LaTeXFormatInclude) $
      do writeFile (outputpath ++ "/tableAbsRelLoggingsSegmentationPerWeek.tex") $
               renderLateX (showAsTable2D $ absRelAnalysis <$> input)
         return ()
  
   return $
                    (section Nothing ! [formatonly $ show HtmlFormat] << (unwords (researchname:[partname])))
--                      +++
--                     locResearchLOC
--                      +++
--                     locResearchLOD
--                      +++
--                     locResearchLOE
--                      +++
--                     locResearchLOD_LOD_Ratio
                     +++
                    plotToFigure' outputpath abslinedistribution
                     +++
                    plotToFigure' outputpath rellinedistribution
                     +++
                    (showAsTable2D $ absRelAnalysis <$> input)

